from .base import Delivery, Sink, Source
from .memory import MemoryQueue
from .mysql import MySQLConnector
from .rabbitmq import RabbitMQConnector
from .schedule import CronSource, IntervalSource
from .sqs import SQSConnector
from .webhook import BearerAuth, WebhookResponse, WebhookSource

__all__ = [
    "BearerAuth",
    "CronSource",
    "Delivery",
    "IntervalSource",
    "MemoryQueue",
    "MySQLConnector",
    "RabbitMQConnector",
    "Sink",
    "Source",
    "SQSConnector",
    "WebhookResponse",
    "WebhookSource",
]



from __future__ import annotations

import asyncio
import logging
import signal

from onestep import (
    ConnectorErrorKind,
    ConnectorOperation,
    ConnectorOperationError,
    FailureKind,
    InMemoryMetrics,
    InMemoryStateStore,
    MaxAttempts,
    MemoryQueue,
    NoRetry,
    OneStepApp,
    RetryAction,
    RetryDecision,
    Sink,
    StructuredEventLogger,
    TaskEventKind,
)
from onestep.connectors.base import Delivery, Sink, Source
from onestep.envelope import Envelope


class _StubDelivery(Delivery):
    def __init__(self, payload):
        super().__init__(Envelope(body=payload))
        self.acked = False

    async def ack(self) -> None:
        self.acked = True

    async def retry(self, *, delay_s: float | None = None) -> None:
        return None

    async def fail(self, exc: Exception | None = None) -> None:
        return None


class _FlakySource(Source):
    def __init__(self, name: str, *, retryable: bool) -> None:
        super().__init__(name)
        self.poll_interval_s = 0.01
        self._retryable = retryable
        self._calls = 0
        self.delivery = _StubDelivery({"value": 1})

    async def fetch(self, limit: int) -> list[Delivery]:
        self._calls += 1
        if self._calls == 1:
            raise ConnectorOperationError(
                backend="fake",
                operation=ConnectorOperation.FETCH,
                kind=ConnectorErrorKind.DISCONNECTED if self._retryable else ConnectorErrorKind.MISCONFIGURED,
                source_name=self.name,
                retry_delay_s=0,
            )
        if self._calls == 2:
            return [self.delivery]
        return []


class _FlakySink(Sink):
    def __init__(self, name: str, *, failures: int, retryable: bool) -> None:
        super().__init__(name)
        self.failures = failures
        self.retryable = retryable
        self.calls = 0
        self.items = []

    async def send(self, envelope: Envelope) -> None:
        self.calls += 1
        if self.calls <= self.failures:
            raise ConnectorOperationError(
                backend="fake",
                operation=ConnectorOperation.SEND,
                kind=ConnectorErrorKind.DISCONNECTED if self.retryable else ConnectorErrorKind.MISCONFIGURED,
                source_name=self.name,
                retry_delay_s=0,
            )
        self.items.append(envelope.body)


class _ManagedSource(Source):
    def __init__(self, name: str, *, fail_open: bool = False) -> None:
        super().__init__(name)
        self.fail_open = fail_open
        self.opened = False
        self.closed = False
        self.fetch_calls = 0

    async def open(self) -> None:
        self.opened = True
        if self.fail_open:
            raise RuntimeError(f"{self.name} failed to open")

    async def fetch(self, limit: int) -> list[Delivery]:
        self.fetch_calls += 1
        return []

    async def close(self) -> None:
        self.closed = True


def test_return_value_publishes_to_default_sink_contract() -> None:
    async def scenario() -> None:
        source = MemoryQueue("incoming")
        sink = MemoryQueue("processed")
        app = OneStepApp("return-contract")

        @app.task(source=source, emit=sink)
        async def transform(ctx, item):
            ctx.app.request_shutdown()
            return {"value": item["value"] * 2}

        await source.publish({"value": 21})
        await app.serve()

        deliveries = await sink.fetch(1)
        assert len(deliveries) == 1
        assert deliveries[0].payload == {"value": 42}

    asyncio.run(scenario())


def test_retryable_source_fetch_error_does_not_exit_worker_contract() -> None:
    async def scenario() -> None:
        source = _FlakySource("flaky-source", retryable=True)
        app = OneStepApp("retryable-source")
        seen: list[dict[str, int]] = []

        @app.task(source=source)
        async def consume(ctx, item):
            seen.append(item)
            ctx.app.request_shutdown()

        await app.serve()

        assert seen == [{"value": 1}]
        assert source.delivery.acked is True

    asyncio.run(scenario())


def test_non_retryable_source_fetch_error_still_fails_fast_contract() -> None:
    async def scenario() -> None:
        source = _FlakySource("broken-source", retryable=False)
        app = OneStepApp("broken-source")

        @app.task(source=source)
        async def consume(ctx, item):
            raise AssertionError("should not run")

        try:
            await app.serve()
        except ConnectorOperationError as exc:
            assert exc.kind is ConnectorErrorKind.MISCONFIGURED
        else:
            raise AssertionError("expected ConnectorOperationError")

    asyncio.run(scenario())


def test_retryable_sink_send_error_retries_and_succeeds_contract() -> None:
    async def scenario() -> None:
        source = MemoryQueue("incoming")
        sink = _FlakySink("flaky-sink", failures=1, retryable=True)
        app = OneStepApp("retryable-sink")

        @app.task(source=source, emit=sink)
        async def consume(ctx, item):
            ctx.app.request_shutdown()
            return {"value": item["value"] + 1}

        await source.publish({"value": 1})
        await app.serve()

        assert sink.calls == 2
        assert sink.items == [{"value": 2}]

    asyncio.run(scenario())


def test_startup_failure_closes_resources_opened_before_failing_resource_contract() -> None:
    async def scenario() -> None:
        healthy = _ManagedSource("healthy")
        broken = _ManagedSource("broken", fail_open=True)
        app = OneStepApp("startup-resource-failure")

        @app.task(source=healthy)
        async def consume_healthy(ctx, item):
            return None

        @app.task(source=broken)
        async def consume_broken(ctx, item):
            return None

        try:
            await app.startup()
        except RuntimeError as exc:
            assert "failed to open" in str(exc)
        else:
            raise AssertionError("expected startup failure")

        assert healthy.opened is True
        assert healthy.closed is True
        assert broken.opened is True
        assert app._resources == []

    asyncio.run(scenario())


def test_startup_hook_failure_closes_opened_resources_contract() -> None:
    async def scenario() -> None:
        source = _ManagedSource("source")
        app = OneStepApp("startup-hook-failure")

        @app.task(source=source)
        async def consume(ctx, item):
            return None

        @app.on_startup
        async def broken_startup(app):
            raise RuntimeError("startup hook broke")

        try:
            await app.startup()
        except RuntimeError as exc:
            assert str(exc) == "startup hook broke"
        else:
            raise AssertionError("expected startup hook failure")

        assert source.opened is True
        assert source.closed is True
        assert app._resources == []

    asyncio.run(scenario())


def test_app_created_outside_running_loop_still_serves_contract() -> None:
    source = MemoryQueue("outside-loop.incoming")
    sink = MemoryQueue("outside-loop.processed")
    app = OneStepApp("outside-loop-app")

    @app.task(source=source, emit=sink)
    async def transform(ctx, item):
        ctx.app.request_shutdown()
        return {"value": item["value"] + 1}

    async def scenario() -> None:
        await source.publish({"value": 1})
        await app.serve()

        deliveries = await sink.fetch(1)
        assert len(deliveries) == 1
        assert deliveries[0].payload == {"value": 2}

    asyncio.run(scenario())


def test_run_installs_sigterm_handler_contract(monkeypatch) -> None:
    app = OneStepApp("signal-contract")
    handlers = {}

    def fake_getsignal(sig):
        return f"previous-{sig}"

    def fake_signal(sig, handler):
        handlers[sig] = handler

    monkeypatch.setattr(signal, "getsignal", fake_getsignal)
    monkeypatch.setattr(signal, "signal", fake_signal)

    with app._install_signal_handlers():
        assert callable(handlers[signal.SIGTERM])
        handlers[signal.SIGTERM](signal.SIGTERM, None)
        assert app.is_stopping is True

    assert handlers[signal.SIGTERM] == f"previous-{signal.SIGTERM}"


def test_ctx_emit_and_return_follow_separate_contracts() -> None:
    async def scenario() -> None:
        source = MemoryQueue("incoming")
        default_sink = MemoryQueue("default")
        explicit_sink = MemoryQueue("explicit")
        app = OneStepApp("emit-contract")

        @app.task(source=source, emit=default_sink)
        async def fanout(ctx, item):
            await ctx.emit({"kind": "side", "value": item["value"]}, sink=explicit_sink)
            ctx.app.request_shutdown()
            return {"kind": "main", "value": item["value"] * 2}

        await source.publish({"value": 3})
        await app.serve()

        default_batch = await default_sink.fetch(1)
        explicit_batch = await explicit_sink.fetch(1)
        assert len(default_batch) == 1
        assert len(explicit_batch) == 1
        assert default_batch[0].payload == {"kind": "main", "value": 6}
        assert explicit_batch[0].payload == {"kind": "side", "value": 3}

    asyncio.run(scenario())


def test_task_timeout_retries_once_then_fails() -> None:
    async def scenario() -> None:
        source = MemoryQueue("incoming", poll_interval_s=0.01)
        app = OneStepApp("timeout-app")
        attempts: list[int] = []

        @app.task(source=source, retry=MaxAttempts(2, delay_s=0), timeout_s=0.01)
        async def slow(ctx, item):
            attempts.append(ctx.current.attempts)
            await asyncio.sleep(0.05)

        await source.publish({"value": 1})
        app_task = asyncio.create_task(app.serve())
        await asyncio.sleep(0.15)
        app.request_shutdown()
        await asyncio.wait_for(app_task, timeout=1.0)

        assert attempts == [0, 1]
        assert await source.fetch(1) == []

    asyncio.run(scenario())


def test_failure_classification_and_dead_letter_payload() -> None:
    class CapturePolicy:
        def __init__(self) -> None:
            self.failures = []

        def on_error(self, envelope, exc, failure):
            self.failures.append(failure)
            return RetryAction(RetryDecision.FAIL)

    async def scenario() -> None:
        source = MemoryQueue("incoming", poll_interval_s=0.01)
        dead = MemoryQueue("dead", poll_interval_s=0.01)
        policy = CapturePolicy()
        app = OneStepApp("dlq-app")

        @app.task(source=source, retry=policy, timeout_s=0.01, dead_letter=dead)
        async def slow(ctx, item):
            await asyncio.sleep(0.05)

        await source.publish({"value": 1}, meta={"trace_id": "abc"})
        app_task = asyncio.create_task(app.serve())
        await asyncio.sleep(0.1)
        app.request_shutdown()
        await asyncio.wait_for(app_task, timeout=1.0)

        assert len(policy.failures) == 1
        assert policy.failures[0].kind is FailureKind.TIMEOUT

        dead_batch = await dead.fetch(1)
        assert len(dead_batch) == 1
        assert dead_batch[0].payload == {
            "payload": {"value": 1},
            "failure": {
                "kind": "timeout",
                "exception_type": "TimeoutError",
                "message": "",
            },
        }
        assert dead_batch[0].envelope.meta == {
            "app": "dlq-app",
            "task": "slow",
            "source": "incoming",
            "original_meta": {"trace_id": "abc"},
            "original_attempts": 0,
        }
        assert await source.fetch(1) == []

    asyncio.run(scenario())


def test_task_events_and_metrics_for_success_and_timeout_dead_letter() -> None:
    async def scenario() -> None:
        source = MemoryQueue("incoming", poll_interval_s=0.01)
        dead = MemoryQueue("dead", poll_interval_s=0.01)
        metrics = InMemoryMetrics()
        events = []
        app = OneStepApp("events-app")
        app.on_event(metrics)

        @app.on_event
        def capture(event):
            events.append(event.kind)

        @app.on_startup
        async def seed(app):
            await source.publish({"kind": "ok"})
            await source.publish({"kind": "slow"})

        @app.task(source=source, retry=NoRetry(), timeout_s=0.01, dead_letter=dead)
        async def consume(ctx, item):
            if item["kind"] == "slow":
                await asyncio.sleep(0.05)
                return None
            if item["kind"] == "ok":
                return {"done": True}
            raise AssertionError("unexpected payload")

        app_task = asyncio.create_task(app.serve())
        await asyncio.sleep(0.1)
        app.request_shutdown()
        await asyncio.wait_for(app_task, timeout=1.0)

        assert metrics.count(TaskEventKind.FETCHED) == 2
        assert metrics.count(TaskEventKind.STARTED) == 2
        assert metrics.count(TaskEventKind.SUCCEEDED) == 1
        assert metrics.count(TaskEventKind.DEAD_LETTERED) == 1
        assert metrics.count(TaskEventKind.FAILED) == 1
        assert metrics.count(TaskEventKind.FAILED, failure_kind=FailureKind.TIMEOUT) == 1
        assert events.count(TaskEventKind.SUCCEEDED) == 1
        assert events.count(TaskEventKind.DEAD_LETTERED) == 1

        dead_batch = await dead.fetch(1)
        assert len(dead_batch) == 1
        assert dead_batch[0].payload["payload"] == {"kind": "slow"}
        assert dead_batch[0].payload["failure"]["kind"] == "timeout"

    asyncio.run(scenario())


def test_structured_event_logger_emits_uniform_log_fields() -> None:
    class ListHandler(logging.Handler):
        def __init__(self) -> None:
            super().__init__()
            self.records = []

        def emit(self, record) -> None:
            self.records.append(record)

    async def scenario() -> None:
        source = MemoryQueue("incoming")
        logger = logging.getLogger("tests.structured-events")
        handler = ListHandler()
        logger.handlers = [handler]
        logger.setLevel(logging.DEBUG)
        logger.propagate = False

        app = OneStepApp("structured-events")
        app.on_event(StructuredEventLogger(logger=logger))

        @app.on_startup
        async def seed(app):
            await source.publish({"value": 4})

        @app.task(source=source)
        async def consume(ctx, item):
            ctx.app.request_shutdown()
            return None

        await asyncio.wait_for(app.serve(), timeout=1.0)

        succeeded = [record for record in handler.records if getattr(record, "event_kind", None) == "succeeded"]
        assert len(succeeded) == 1
        record = succeeded[0]
        assert record.levelno == logging.INFO
        assert record.app_name == "structured-events"
        assert record.task_name == "consume"
        assert record.source_name == "incoming"
        assert record.attempts == 0
        assert isinstance(record.task_event_meta, dict)
        assert record.failure_kind is None

    asyncio.run(scenario())


def test_structured_event_logger_includes_failure_fields() -> None:
    class ListHandler(logging.Handler):
        def __init__(self) -> None:
            super().__init__()
            self.records = []

        def emit(self, record) -> None:
            self.records.append(record)

    async def scenario() -> None:
        source = MemoryQueue("incoming", poll_interval_s=0.01)
        logger = logging.getLogger("tests.structured-events.failure")
        handler = ListHandler()
        logger.handlers = [handler]
        logger.setLevel(logging.DEBUG)
        logger.propagate = False

        app = OneStepApp("structured-events-failure")
        app.on_event(StructuredEventLogger(logger=logger))

        @app.on_startup
        async def seed(app):
            await source.publish({"value": 1})

        @app.task(source=source, retry=NoRetry(), timeout_s=0.01)
        async def consume(ctx, item):
            await asyncio.sleep(0.05)

        app_task = asyncio.create_task(app.serve())
        await asyncio.sleep(0.1)
        app.request_shutdown()
        await asyncio.wait_for(app_task, timeout=1.0)

        failed = [record for record in handler.records if getattr(record, "event_kind", None) == "failed"]
        assert len(failed) == 1
        record = failed[0]
        assert record.levelno == logging.ERROR
        assert record.failure_kind == "timeout"
        assert record.failure_exception_type == "TimeoutError"
        assert record.app_name == "structured-events-failure"
        assert record.task_name == "consume"

    asyncio.run(scenario())


def test_shutdown_timeout_cancels_inflight_and_retries_delivery() -> None:
    async def scenario() -> None:
        source = MemoryQueue("incoming", poll_interval_s=0.01)
        app = OneStepApp("shutdown-app", shutdown_timeout_s=0.05)
        started = asyncio.Event()
        cancelled = asyncio.Event()

        @app.task(source=source)
        async def slow(ctx, item):
            started.set()
            try:
                await asyncio.sleep(10)
            finally:
                cancelled.set()

        await source.publish({"value": 1})
        app_task = asyncio.create_task(app.serve())
        await asyncio.wait_for(started.wait(), timeout=1.0)
        app.request_shutdown()
        await asyncio.wait_for(app_task, timeout=1.0)

        assert cancelled.is_set()
        redelivered = await source.fetch(1)
        assert len(redelivered) == 1
        assert redelivered[0].payload == {"value": 1}
        assert redelivered[0].envelope.attempts == 1

    asyncio.run(scenario())


def test_event_handler_failure_does_not_break_runtime() -> None:
    async def scenario() -> None:
        source = MemoryQueue("incoming")
        app = OneStepApp("event-handler-failure")
        seen = []

        @app.on_event
        def broken(event):
            raise RuntimeError("observer broke")

        @app.on_event
        def capture(event):
            seen.append(event.kind)

        @app.on_startup
        async def seed(app):
            await source.publish({"value": 1})

        @app.task(source=source)
        async def consume(ctx, item):
            ctx.app.request_shutdown()
            return None

        await asyncio.wait_for(app.serve(), timeout=1.0)
        assert TaskEventKind.SUCCEEDED in seen

    asyncio.run(scenario())


def test_dead_letter_publish_failure_retries_original_delivery() -> None:
    class BrokenSink(Sink):
        def __init__(self) -> None:
            super().__init__("broken")

        async def send(self, envelope) -> None:
            raise RuntimeError("dead-letter offline")

    async def scenario() -> None:
        source = MemoryQueue("incoming", poll_interval_s=0.01)
        app = OneStepApp("broken-dlq")

        @app.task(source=source, retry=NoRetry(), dead_letter=BrokenSink())
        async def fail_now(ctx, item):
            ctx.app.request_shutdown()
            raise RuntimeError("boom")

        await source.publish({"value": 1})
        await asyncio.wait_for(app.serve(), timeout=1.0)

        redelivered = await source.fetch(1)
        assert len(redelivered) == 1
        assert redelivered[0].payload == {"value": 1}
        assert redelivered[0].envelope.attempts == 1

    asyncio.run(scenario())


def test_lifecycle_hooks_and_context_state_config() -> None:
    async def scenario() -> None:
        source = MemoryQueue("incoming")
        state = InMemoryStateStore()
        app = OneStepApp("hooked-app", config={"prefix": "v1"}, state=state)
        events: list[str] = []

        @app.on_startup
        async def startup(app):
            events.append("startup")
            await source.publish({"value": 2})

        @app.on_shutdown
        def shutdown():
            events.append("shutdown")

        @app.task(source=source)
        async def consume(ctx, item):
            events.append("handler")
            assert ctx.config["prefix"] == "v1"
            current = await ctx.state.get("count", 0)
            await ctx.state.set("count", current + item["value"])
            ctx.app.request_shutdown()

        await app.serve()

        assert events == ["startup", "handler", "shutdown"]
        assert await state.load("hooked-app:consume:count") == 2

    asyncio.run(scenario())



from onestep import MemoryQueue, MySQLConnector, OneStepApp

app = OneStepApp("sync-users")
db = MySQLConnector("mysql+pymysql://root:root@localhost:3306/app")
state = db.cursor_store(table="onestep_cursor")
source = db.incremental(
    table="users",
    key="id",
    cursor=("updated_at", "id"),
    where="deleted = 0",
    batch_size=1000,
    state=state,
)
out = MemoryQueue("dw-users")


@app.task(source=source, emit=out, concurrency=1)
async def sync_user(ctx, row):
    return {"id": row["id"], "name": row["name"], "updated_at": row["updated_at"]}



from __future__ import annotations

import asyncio
import hashlib
import json
import logging
import math
import os
import platform
import re
import socket
from dataclasses import dataclass, field
from datetime import datetime, timezone
from importlib.metadata import PackageNotFoundError, version as package_version
from typing import TYPE_CHECKING, Any, Awaitable, Callable
from urllib import error, request
from uuid import UUID, uuid4

from .events import TaskEvent, TaskEventKind
from .retry import FailureKind

if TYPE_CHECKING:
    from .app import OneStepApp


def _utcnow() -> datetime:
    return datetime.now(timezone.utc)


def _resolve_onestep_version() -> str:
    try:
        return package_version("onestep")
    except PackageNotFoundError:  # pragma: no cover - local source tree before install
        return "dev"


def _json_default(value: Any) -> Any:
    if isinstance(value, datetime):
        return value.isoformat()
    if isinstance(value, UUID):
        return str(value)
    raise TypeError(f"Object of type {type(value).__name__} is not JSON serializable")


def _normalize_base_url(value: str) -> str:
    normalized = value.strip().rstrip("/")
    if not normalized:
        raise ValueError("control plane base URL must not be empty")
    return normalized


def _coerce_positive_float(name: str, value: float) -> float:
    if value <= 0:
        raise ValueError(f"{name} must be > 0")
    return value


def _coerce_positive_int(name: str, value: int) -> int:
    if value < 1:
        raise ValueError(f"{name} must be >= 1")
    return value


def _percentile(values: list[float], percentile: float) -> float | None:
    if not values:
        return None
    ordered = sorted(values)
    index = max(0, math.ceil(percentile * len(ordered)) - 1)
    return ordered[index]


def _read_env(*names: str) -> str | None:
    for name in names:
        value = os.getenv(name)
        if value is not None and value.strip():
            return value.strip()
    return None


_KIND_OVERRIDES = {
    "CronSource": "cron",
    "IncrementalTableSource": "mysql_incremental",
    "IntervalSource": "interval",
    "MaxAttempts": "max_attempts",
    "MemoryQueue": "memory_queue",
    "NoRetry": "no_retry",
    "RabbitMQQueue": "rabbitmq_queue",
    "SQSQueue": "sqs_queue",
    "TableQueueSource": "mysql_table_queue",
    "TableSink": "mysql_table_sink",
    "WebhookSource": "webhook",
}


def _snake_case(value: str) -> str:
    normalized = re.sub(r"(?<!^)(?=[A-Z])", "_", value).lower()
    return normalized.strip("_")


def _kind_name(value: object) -> str:
    class_name = value.__class__.__name__
    if class_name in _KIND_OVERRIDES:
        return _KIND_OVERRIDES[class_name]
    for suffix in ("Source", "Sink"):
        if class_name.endswith(suffix):
            class_name = class_name[: -len(suffix)]
            break
    return _snake_case(class_name)


def _json_safe(value: Any) -> Any:
    if value is None or isinstance(value, (str, int, float, bool)):
        return value
    if isinstance(value, datetime):
        return value.isoformat()
    if isinstance(value, UUID):
        return str(value)
    if isinstance(value, tuple):
        return [_json_safe(item) for item in value]
    if isinstance(value, list):
        return [_json_safe(item) for item in value]
    if isinstance(value, dict):
        return {str(key): _json_safe(item) for key, item in value.items()}
    enum_value = getattr(value, "value", None)
    if isinstance(enum_value, (str, int, float, bool)):
        return enum_value
    return str(value)


def _stable_hash(payload: dict[str, Any]) -> str:
    encoded = json.dumps(
        payload,
        sort_keys=True,
        separators=(",", ":"),
        default=_json_default,
    ).encode("utf-8")
    return f"sha256:{hashlib.sha256(encoded).hexdigest()}"


def _seconds_value(value: float) -> int | float:
    if float(value).is_integer():
        return int(value)
    return value


@dataclass
class ControlPlaneReporterConfig:
    base_url: str
    token: str
    environment: str = "dev"
    service_name: str | None = None
    node_name: str | None = None
    deployment_version: str | None = None
    instance_id: UUID = field(default_factory=uuid4)
    heartbeat_interval_s: float = 30.0
    metrics_interval_s: float = 30.0
    event_flush_interval_s: float = 5.0
    event_batch_size: int = 100
    timeout_s: float = 5.0

    def __post_init__(self) -> None:
        self.base_url = _normalize_base_url(self.base_url)
        self.token = self.token.strip()
        if not self.token:
            raise ValueError("control plane token must not be empty")
        if self.environment not in {"dev", "staging", "prod"}:
            raise ValueError("environment must be one of: dev, staging, prod")
        self.heartbeat_interval_s = _coerce_positive_float(
            "heartbeat_interval_s",
            self.heartbeat_interval_s,
        )
        self.metrics_interval_s = _coerce_positive_float(
            "metrics_interval_s",
            self.metrics_interval_s,
        )
        self.event_flush_interval_s = _coerce_positive_float(
            "event_flush_interval_s",
            self.event_flush_interval_s,
        )
        self.event_batch_size = _coerce_positive_int("event_batch_size", self.event_batch_size)
        self.timeout_s = _coerce_positive_float("timeout_s", self.timeout_s)

    @classmethod
    def from_env(cls, *, app_name: str | None = None) -> "ControlPlaneReporterConfig":
        base_url = _read_env("ONESTEP_CONTROL_PLANE_URL", "ONESTEP_CONTROL_URL")
        token = _read_env("ONESTEP_CONTROL_PLANE_TOKEN", "ONESTEP_CONTROL_TOKEN")
        if base_url is None:
            raise ValueError("missing ONESTEP_CONTROL_PLANE_URL or ONESTEP_CONTROL_URL")
        if token is None:
            raise ValueError("missing ONESTEP_CONTROL_PLANE_TOKEN or ONESTEP_CONTROL_TOKEN")
        environment = _read_env("ONESTEP_CONTROL_PLANE_ENVIRONMENT", "ONESTEP_ENV") or "dev"
        service_name = _read_env("ONESTEP_SERVICE_NAME") or app_name
        node_name = _read_env("ONESTEP_NODE_NAME")
        deployment_version = _read_env(
            "ONESTEP_DEPLOYMENT_VERSION",
            "ONESTEP_VERSION",
        )
        instance_id_value = _read_env("ONESTEP_INSTANCE_ID")
        instance_id = UUID(instance_id_value) if instance_id_value is not None else uuid4()
        heartbeat_interval_s = float(_read_env("ONESTEP_CONTROL_PLANE_HEARTBEAT_INTERVAL_S") or 30.0)
        metrics_interval_s = float(_read_env("ONESTEP_CONTROL_PLANE_METRICS_INTERVAL_S") or 30.0)
        event_flush_interval_s = float(
            _read_env("ONESTEP_CONTROL_PLANE_EVENT_FLUSH_INTERVAL_S") or 5.0
        )
        event_batch_size = int(_read_env("ONESTEP_CONTROL_PLANE_EVENT_BATCH_SIZE") or 100)
        timeout_s = float(_read_env("ONESTEP_CONTROL_PLANE_TIMEOUT_S") or 5.0)
        return cls(
            base_url=base_url,
            token=token,
            environment=environment,
            service_name=service_name,
            node_name=node_name,
            deployment_version=deployment_version,
            instance_id=instance_id,
            heartbeat_interval_s=heartbeat_interval_s,
            metrics_interval_s=metrics_interval_s,
            event_flush_interval_s=event_flush_interval_s,
            event_batch_size=event_batch_size,
            timeout_s=timeout_s,
        )


@dataclass
class _TaskMetricsState:
    fetched: int = 0
    started: int = 0
    succeeded: int = 0
    retried: int = 0
    failed: int = 0
    dead_lettered: int = 0
    cancelled: int = 0
    timeouts: int = 0
    inflight: int = 0
    durations_ms: list[float] = field(default_factory=list)

    def has_activity(self) -> bool:
        return any(
            (
                self.fetched,
                self.started,
                self.succeeded,
                self.retried,
                self.failed,
                self.dead_lettered,
                self.cancelled,
                self.timeouts,
                self.inflight,
            )
        )

    def rotate(self) -> "_TaskMetricsState":
        rotated = _TaskMetricsState(
            fetched=self.fetched,
            started=self.started,
            succeeded=self.succeeded,
            retried=self.retried,
            failed=self.failed,
            dead_lettered=self.dead_lettered,
            cancelled=self.cancelled,
            timeouts=self.timeouts,
            inflight=self.inflight,
            durations_ms=list(self.durations_ms),
        )
        self.fetched = 0
        self.started = 0
        self.succeeded = 0
        self.retried = 0
        self.failed = 0
        self.dead_lettered = 0
        self.cancelled = 0
        self.timeouts = 0
        self.durations_ms.clear()
        return rotated


@dataclass
class _PreparedMetricsBatch:
    window_started_at: datetime
    window_ended_at: datetime
    tasks: list[dict[str, Any]]


ReporterSender = Callable[[str, dict[str, Any]], Awaitable[None]]


class ControlPlaneReporter:
    def __init__(
        self,
        config: ControlPlaneReporterConfig,
        *,
        sender: ReporterSender | None = None,
    ) -> None:
        self.config = config
        self._sender = sender or self._default_sender
        self._app: OneStepApp | None = None
        self._logger = logging.getLogger("onestep.control_plane")
        self._attached = False
        self._started_at = _utcnow()
        self._hostname = socket.gethostname()
        self._node_name = config.node_name or self._hostname
        self._service_name = config.service_name
        self._deployment_version = config.deployment_version or _resolve_onestep_version()
        self._onestep_version = _resolve_onestep_version()
        self._sequence = 0
        self._event_sequence = 0
        self._pending_events: list[dict[str, Any]] = []
        self._metrics_window_started_at = _utcnow()
        self._metrics_by_task: dict[str, _TaskMetricsState] = {}
        self._pending_metric_batches: list[_PreparedMetricsBatch] = []
        self._degraded_since_last_heartbeat = False
        self._last_synced_topology_hash: str | None = None
        self._loop_tasks: list[asyncio.Task[None]] = []
        self._stop_event: asyncio.Event | None = None
        self._event_flush_signal: asyncio.Event | None = None

    def attach(self, app: "OneStepApp") -> "ControlPlaneReporter":
        if self._attached:
            if self._app is not app:
                raise RuntimeError("reporter is already attached to another app")
            return self
        self._attached = True
        self._app = app
        if self._service_name is None:
            self._service_name = app.name
        app.on_startup(self.startup)
        app.on_shutdown(self.shutdown)
        app.on_event(self.handle_event)
        return self

    async def startup(self) -> None:
        self._sequence = 0
        self._event_sequence = 0
        self._pending_events.clear()
        self._pending_metric_batches.clear()
        self._metrics_by_task.clear()
        self._degraded_since_last_heartbeat = False
        self._last_synced_topology_hash = None
        self._stop_event = asyncio.Event()
        self._event_flush_signal = asyncio.Event()
        self._started_at = _utcnow()
        self._metrics_window_started_at = self._started_at
        await self._safe_send_heartbeat()
        await self._safe_send_sync()
        self._loop_tasks = [
            asyncio.create_task(self._heartbeat_loop(), name="onestep-control-plane-heartbeat"),
            asyncio.create_task(self._metrics_loop(), name="onestep-control-plane-metrics"),
            asyncio.create_task(self._events_loop(), name="onestep-control-plane-events"),
        ]

    async def shutdown(self) -> None:
        if self._stop_event is not None:
            self._stop_event.set()
        if self._event_flush_signal is not None:
            self._event_flush_signal.set()
        if self._loop_tasks:
            await asyncio.gather(*self._loop_tasks, return_exceptions=True)
        self._loop_tasks = []
        self._rotate_metrics_window(_utcnow())
        await self._flush_metric_batches()
        await self._flush_event_batches()

    def handle_event(self, event: TaskEvent) -> None:
        bucket = self._metrics_by_task.setdefault(event.task, _TaskMetricsState())
        if event.kind is TaskEventKind.FETCHED:
            bucket.fetched += 1
        elif event.kind is TaskEventKind.STARTED:
            bucket.started += 1
            bucket.inflight += 1
        elif event.kind is TaskEventKind.SUCCEEDED:
            bucket.succeeded += 1
            bucket.inflight = max(0, bucket.inflight - 1)
            self._record_duration(bucket, event)
        elif event.kind is TaskEventKind.RETRIED:
            bucket.retried += 1
            bucket.inflight = max(0, bucket.inflight - 1)
            self._record_timeout(bucket, event)
            self._record_duration(bucket, event)
            self._degraded_since_last_heartbeat = True
        elif event.kind is TaskEventKind.FAILED:
            bucket.failed += 1
            bucket.inflight = max(0, bucket.inflight - 1)
            self._record_timeout(bucket, event)
            self._record_duration(bucket, event)
            self._degraded_since_last_heartbeat = True
        elif event.kind is TaskEventKind.DEAD_LETTERED:
            bucket.dead_lettered += 1
            self._degraded_since_last_heartbeat = True
        elif event.kind is TaskEventKind.CANCELLED:
            bucket.cancelled += 1
            self._degraded_since_last_heartbeat = True

        if event.kind in {
            TaskEventKind.RETRIED,
            TaskEventKind.FAILED,
            TaskEventKind.DEAD_LETTERED,
            TaskEventKind.CANCELLED,
        }:
            self._pending_events.append(self._build_event_payload(event))
            if (
                self._event_flush_signal is not None
                and len(self._pending_events) >= self.config.event_batch_size
            ):
                self._event_flush_signal.set()

    async def _heartbeat_loop(self) -> None:
        while not await self._wait_for_stop(self.config.heartbeat_interval_s):
            await self._safe_send_sync()
            await self._safe_send_heartbeat()

    async def _metrics_loop(self) -> None:
        while not await self._wait_for_stop(self.config.metrics_interval_s):
            self._rotate_metrics_window(_utcnow())
            await self._flush_metric_batches()

    async def _events_loop(self) -> None:
        while True:
            if self._stop_event is None or self._event_flush_signal is None:
                return
            if self._stop_event.is_set():
                return
            try:
                await asyncio.wait_for(
                    self._event_flush_signal.wait(),
                    timeout=self.config.event_flush_interval_s,
                )
            except asyncio.TimeoutError:
                pass
            else:
                self._event_flush_signal.clear()
            if self._stop_event.is_set():
                return
            await self._flush_event_batches()

    async def _wait_for_stop(self, timeout_s: float) -> bool:
        if self._stop_event is None:
            return True
        try:
            await asyncio.wait_for(self._stop_event.wait(), timeout=timeout_s)
        except asyncio.TimeoutError:
            return False
        return True

    def _next_sequence(self) -> int:
        self._sequence += 1
        return self._sequence

    def _next_event_id(self) -> str:
        self._event_sequence += 1
        return f"{self.config.instance_id.hex}-{self._event_sequence}"

    def _service_descriptor(self) -> dict[str, Any]:
        return {
            "name": self._service_name,
            "environment": self.config.environment,
            "node_name": self._node_name,
            "instance_id": self.config.instance_id,
            "deployment_version": self._deployment_version,
        }

    def _build_runtime_descriptor(self) -> dict[str, Any]:
        return {
            "onestep_version": self._onestep_version,
            "python_version": platform.python_version(),
            "hostname": self._hostname,
            "pid": os.getpid(),
            "started_at": self._started_at,
        }

    def _build_health_descriptor(self) -> dict[str, Any]:
        inflight_tasks = sum(bucket.inflight for bucket in self._metrics_by_task.values())
        status = "degraded" if self._degraded_since_last_heartbeat else "ok"
        return {
            "status": status,
            "uptime_s": max(0, int((_utcnow() - self._started_at).total_seconds())),
            "inflight_tasks": inflight_tasks,
        }

    def _build_retry_descriptor(self, policy: Any) -> dict[str, Any]:
        kind = _kind_name(policy)
        if kind == "max_attempts":
            return {
                "kind": kind,
                "config": {
                    "max_attempts": policy.max_attempts,
                    "delay_s": policy.delay_s,
                },
            }
        return {"kind": kind, "config": {}}

    def _build_connector_config(self, resource: Any) -> dict[str, Any]:
        class_name = resource.__class__.__name__
        if class_name == "IntervalSource":
            return {
                "seconds": _seconds_value(resource.interval.total_seconds()),
                "immediate": resource.immediate,
                "overlap": resource.overlap,
                "timezone": str(resource.timezone),
                "poll_interval_s": resource.poll_interval_s,
            }
        if class_name == "CronSource":
            return {
                "expression": resource.schedule.expression,
                "immediate": resource.immediate,
                "overlap": resource.overlap,
                "timezone": str(resource.timezone),
                "poll_interval_s": resource.poll_interval_s,
            }
        if class_name == "MemoryQueue":
            return {
                "maxsize": resource._maxsize,
                "batch_size": resource.batch_size,
                "poll_interval_s": resource.poll_interval_s,
            }
        if class_name == "WebhookSource":
            return {
                "path": resource.path,
                "methods": list(resource.methods),
                "host": resource.host,
                "port": resource.port,
                "parser": resource.parser,
                "max_body_bytes": resource.max_body_bytes,
                "read_timeout_s": resource.read_timeout_s,
                "queue_maxsize": resource.queue_maxsize,
                "batch_size": resource.batch_size,
                "poll_interval_s": resource.poll_interval_s,
            }
        if class_name == "RabbitMQQueue":
            return {
                "routing_key": resource.routing_key,
                "exchange": resource.exchange_name,
                "exchange_type": resource.exchange_type,
                "bind": resource.bind,
                "bind_arguments": resource.bind_arguments,
                "durable": resource.durable,
                "auto_delete": resource.auto_delete,
                "arguments": resource.arguments,
                "exchange_durable": resource.exchange_durable,
                "exchange_auto_delete": resource.exchange_auto_delete,
                "exchange_arguments": resource.exchange_arguments,
                "prefetch": resource.prefetch,
                "batch_size": resource.batch_size,
                "poll_interval_s": resource.poll_interval_s,
                "publisher_confirms": resource.publisher_confirms,
                "persistent": resource.persistent,
            }
        if class_name == "SQSQueue":
            return {
                "url": resource.url,
                "wait_time_s": resource.wait_time_s,
                "visibility_timeout": resource.visibility_timeout,
                "batch_size": resource.batch_size,
                "poll_interval_s": resource.poll_interval_s,
                "message_group_id": resource.message_group_id,
                "on_fail": resource.on_fail,
                "delete_batch_size": resource.delete_batch_size,
                "delete_flush_interval_s": resource.delete_flush_interval_s,
                "heartbeat_interval_s": resource.heartbeat_interval_s,
                "heartbeat_visibility_timeout": resource.heartbeat_visibility_timeout,
            }
        if class_name == "TableQueueSource":
            return {
                "table": resource.table_name,
                "key": resource.key,
                "where": resource.where,
                "claim": resource.claim,
                "ack": resource.ack,
                "nack": resource.nack,
                "batch_size": resource.batch_size,
                "poll_interval_s": resource.poll_interval_s,
            }
        if class_name == "IncrementalTableSource":
            return {
                "table": resource.table_name,
                "key": resource.key,
                "cursor": list(resource.cursor),
                "where": resource.where,
                "batch_size": resource.batch_size,
                "poll_interval_s": resource.poll_interval_s,
                "state_key": resource.state_key,
            }
        if class_name == "TableSink":
            return {
                "table": resource.table_name,
                "mode": resource.mode,
                "keys": list(resource.keys),
            }
        return {}

    def _build_connector_descriptor(self, resource: Any) -> dict[str, Any]:
        custom_descriptor = getattr(resource, "control_plane_descriptor", None)
        if callable(custom_descriptor):
            payload = _json_safe(custom_descriptor())
            if isinstance(payload, dict):
                return {
                    "kind": str(payload.get("kind") or _kind_name(resource)),
                    "name": str(payload.get("name") or resource.name),
                    "config": _json_safe(payload.get("config") or {}),
                }
        return {
            "kind": _kind_name(resource),
            "name": resource.name,
            "config": _json_safe(self._build_connector_config(resource)),
        }

    def _build_app_topology_descriptor(self) -> dict[str, Any]:
        if self._app is None:
            raise RuntimeError("reporter is not attached to an app")
        tasks = []
        for task in self._app.tasks:
            tasks.append(
                {
                    "name": task.name,
                    "description": task.description,
                    "source": (
                        self._build_connector_descriptor(task.source) if task.source is not None else None
                    ),
                    "emit": [self._build_connector_descriptor(sink) for sink in task.sinks],
                    "concurrency": task.concurrency,
                    "timeout_s": task.timeout_s,
                    "retry": self._build_retry_descriptor(task.retry),
                }
            )
        return {
            "name": self._service_name,
            "shutdown_timeout_s": self._app.shutdown_timeout_s,
            "tasks": tasks,
        }

    async def _safe_send_sync(self) -> None:
        if self._app is None:
            return
        app_payload = self._build_app_topology_descriptor()
        topology_hash = _stable_hash(app_payload)
        if topology_hash == self._last_synced_topology_hash:
            return
        payload = {
            "service": self._service_descriptor(),
            "runtime": self._build_runtime_descriptor(),
            "app": {
                **app_payload,
                "topology_hash": topology_hash,
            },
            "sent_at": _utcnow(),
            "sequence": self._next_sequence(),
        }
        try:
            await self._sender("/api/v1/agents/sync", payload)
        except Exception:
            self._logger.exception("control plane sync failed")
        else:
            self._last_synced_topology_hash = topology_hash

    async def _safe_send_heartbeat(self) -> None:
        payload = {
            "service": self._service_descriptor(),
            "runtime": self._build_runtime_descriptor(),
            "health": self._build_health_descriptor(),
            "sent_at": _utcnow(),
            "sequence": self._next_sequence(),
        }
        try:
            await self._sender("/api/v1/agents/heartbeat", payload)
        except Exception:
            self._logger.exception("control plane heartbeat failed")
        else:
            self._degraded_since_last_heartbeat = False

    def _rotate_metrics_window(self, ended_at: datetime) -> None:
        tasks: list[dict[str, Any]] = []
        for task_name, bucket in self._metrics_by_task.items():
            if not bucket.has_activity():
                continue
            snapshot = bucket.rotate()
            duration_values = snapshot.durations_ms
            avg_duration_ms = (
                sum(duration_values) / len(duration_values) if duration_values else None
            )
            tasks.append(
                {
                    "task_name": task_name,
                    "window_id": (
                        f"{task_name}:{self._metrics_window_started_at.isoformat()}:{ended_at.isoformat()}"
                    ),
                    "fetched": snapshot.fetched,
                    "started": snapshot.started,
                    "succeeded": snapshot.succeeded,
                    "retried": snapshot.retried,
                    "failed": snapshot.failed,
                    "dead_lettered": snapshot.dead_lettered,
                    "cancelled": snapshot.cancelled,
                    "timeouts": snapshot.timeouts,
                    "inflight": snapshot.inflight,
                    "avg_duration_ms": avg_duration_ms,
                    "p95_duration_ms": _percentile(duration_values, 0.95),
                }
            )
        if tasks:
            self._pending_metric_batches.append(
                _PreparedMetricsBatch(
                    window_started_at=self._metrics_window_started_at,
                    window_ended_at=ended_at,
                    tasks=tasks,
                )
            )
        self._metrics_window_started_at = ended_at

    async def _flush_metric_batches(self) -> None:
        while self._pending_metric_batches:
            batch = self._pending_metric_batches[0]
            payload = {
                "service": self._service_descriptor(),
                "sent_at": _utcnow(),
                "sequence": self._next_sequence(),
                "window": {
                    "started_at": batch.window_started_at,
                    "ended_at": batch.window_ended_at,
                },
                "tasks": batch.tasks,
            }
            try:
                await self._sender("/api/v1/agents/metrics", payload)
            except Exception:
                self._logger.exception("control plane metrics flush failed")
                return
            self._pending_metric_batches.pop(0)

    async def _flush_event_batches(self) -> None:
        while self._pending_events:
            batch = self._pending_events[: self.config.event_batch_size]
            payload = {
                "service": self._service_descriptor(),
                "sent_at": _utcnow(),
                "sequence": self._next_sequence(),
                "events": batch,
            }
            try:
                await self._sender("/api/v1/agents/events", payload)
            except Exception:
                self._logger.exception("control plane event flush failed")
                return
            del self._pending_events[: len(batch)]

    def _build_event_payload(self, event: TaskEvent) -> dict[str, Any]:
        meta = dict(event.meta)
        if event.source is not None:
            meta.setdefault("source", event.source)
        payload: dict[str, Any] = {
            "event_id": self._next_event_id(),
            "kind": event.kind.value,
            "task_name": event.task,
            "occurred_at": event.emitted_at,
            "attempts": event.attempts,
            "duration_ms": (
                int(round(event.duration_s * 1000)) if event.duration_s is not None else None
            ),
            "meta": meta,
        }
        if event.failure is not None:
            payload["failure"] = {
                "kind": event.failure.kind.value,
                "exception_type": event.failure.exception_type,
                "message": event.failure.message,
            }
        return payload

    def _record_duration(self, bucket: _TaskMetricsState, event: TaskEvent) -> None:
        if event.duration_s is None:
            return
        bucket.durations_ms.append(event.duration_s * 1000.0)

    def _record_timeout(self, bucket: _TaskMetricsState, event: TaskEvent) -> None:
        if event.failure is not None and event.failure.kind is FailureKind.TIMEOUT:
            bucket.timeouts += 1

    async def _default_sender(self, endpoint: str, payload: dict[str, Any]) -> None:
        url = f"{self.config.base_url}{endpoint}"
        body = json.dumps(payload, default=_json_default).encode("utf-8")
        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.config.token}",
        }
        await asyncio.to_thread(self._send_request, url, body, headers)

    def _send_request(self, url: str, body: bytes, headers: dict[str, str]) -> None:
        req = request.Request(url, data=body, headers=headers, method="POST")
        try:
            with request.urlopen(req, timeout=self.config.timeout_s) as response:
                status_code = getattr(response, "status", response.getcode())
                if status_code < 200 or status_code >= 300:
                    raise RuntimeError(f"unexpected status code {status_code} from control plane")
                response.read()
        except error.HTTPError as exc:  # pragma: no cover - exercised via default sender integration
            body_text = exc.read().decode("utf-8", errors="replace")
            raise RuntimeError(
                f"control plane request failed with status {exc.code}: {body_text}"
            ) from exc
        except error.URLError as exc:  # pragma: no cover - exercised via default sender integration
            raise RuntimeError(f"control plane request failed: {exc.reason}") from exc


__all__ = ["ControlPlaneReporter", "ControlPlaneReporterConfig"]



from __future__ import annotations

import copy
import importlib
import inspect
import os
from collections.abc import Callable, Mapping, Sequence
from typing import Any

from .app import OneStepApp
from .connectors.base import Sink, Source
from .connectors.memory import MemoryQueue
from .connectors.mysql import MySQLConnector
from .connectors.rabbitmq import RabbitMQConnector
from .connectors.schedule import CronSource, IntervalSource
from .connectors.sqs import SQSConnector
from .connectors.webhook import BearerAuth, WebhookResponse, WebhookSource
from .retry import MaxAttempts, NoRetry, RetryPolicy

_YAML_SUFFIXES = (".yaml", ".yml")


def is_yaml_target(target: str) -> bool:
    return target.lower().endswith(_YAML_SUFFIXES)


def load_yaml_app(path: str) -> OneStepApp:
    yaml = _import_yaml()
    resolved_path = os.path.abspath(path)
    with open(resolved_path, "r", encoding="utf-8") as handle:
        loaded = yaml.safe_load(handle) or {}
    if not isinstance(loaded, Mapping):
        raise TypeError("YAML app config must be a mapping at the top level")
    return load_app_config(loaded, source_path=resolved_path)


def load_app_config(config: Mapping[str, Any], *, source_path: str | None = None) -> OneStepApp:
    app_section = config.get("app")
    if app_section is None:
        app_name = _require_string(config, "name")
        shutdown_timeout_s = config.get("shutdown_timeout_s", 30.0)
        app_config = config.get("config")
        app_state = config.get("state")
    else:
        if not isinstance(app_section, Mapping):
            raise TypeError("'app' must be a mapping")
        app_name = _require_string(app_section, "name")
        shutdown_timeout_s = app_section.get("shutdown_timeout_s", 30.0)
        app_config = app_section.get("config")
        app_state = app_section.get("state")

    if app_config is None:
        app_config = {}
    if not isinstance(app_config, Mapping):
        raise TypeError("'config' must be a mapping when provided")

    app = OneStepApp(
        app_name,
        config=dict(app_config),
        shutdown_timeout_s=shutdown_timeout_s,
    )
    if source_path is not None:
        app.config.setdefault("config_path", source_path)

    resources = _build_resources(config)
    if app_state is not None:
        app.state = _resolve_app_state(resources, app_state)
    tasks = config.get("tasks", [])
    if not isinstance(tasks, Sequence) or isinstance(tasks, (str, bytes)):
        raise TypeError("'tasks' must be a list")

    for index, task_config in enumerate(tasks):
        if not isinstance(task_config, Mapping):
            raise TypeError(f"'tasks[{index}]' must be a mapping")
        task_name = _optional_string(task_config, "name")
        handler = _build_handler(task_config.get("handler"), task_name=task_name)
        app.task(
            name=task_name,
            description=_optional_string(task_config, "description"),
            source=_resolve_optional_source(resources, task_config.get("source"), task_index=index),
            emit=_resolve_optional_sinks(resources, task_config.get("emit"), field="emit", task_index=index),
            dead_letter=_resolve_optional_sinks(
                resources,
                task_config.get("dead_letter"),
                field="dead_letter",
                task_index=index,
            ),
            concurrency=task_config.get("concurrency", 1),
            retry=_build_retry(task_config.get("retry")),
            timeout_s=task_config.get("timeout_s"),
        )(handler)

    return app


def _build_resources(config: Mapping[str, Any]) -> dict[str, Any]:
    specs = _collect_resource_specs(config)
    resources: dict[str, Any] = {}
    resolving: list[str] = []

    def resolve(name: str) -> Any:
        if name in resources:
            return resources[name]
        try:
            spec = specs[name]
        except KeyError as exc:
            raise KeyError(f"unknown resource {name!r}") from exc
        if name in resolving:
            cycle = " -> ".join([*resolving, name])
            raise ValueError(f"cyclic resource reference detected: {cycle}")
        resolving.append(name)
        try:
            built = _build_resource(name, spec, resolve=resolve)
            resources[name] = built
            return built
        finally:
            resolving.pop()

    for name in specs:
        resolve(name)
    return resources


def _collect_resource_specs(config: Mapping[str, Any]) -> dict[str, dict[str, Any]]:
    specs: dict[str, dict[str, Any]] = {}
    for section_name in ("connectors", "sources", "sinks"):
        section = config.get(section_name, {})
        if section is None:
            continue
        if not isinstance(section, Mapping):
            raise TypeError(f"'{section_name}' must be a mapping")
        for name, raw_spec in section.items():
            if not isinstance(raw_spec, Mapping):
                raise TypeError(f"'{section_name}.{name}' must be a mapping")
            spec = dict(raw_spec)
            if name in specs:
                if specs[name] != spec:
                    raise ValueError(f"resource {name!r} is defined multiple times with different configs")
                continue
            specs[name] = spec
    return specs


def _build_resource(name: str, spec: Mapping[str, Any], *, resolve: Callable[[str], Any]) -> Any:
    resource_type = _require_string(spec, "type").strip().lower().replace("-", "_").replace(".", "_")
    resource_name = _resource_name(spec, fallback=name)

    if resource_type == "memory":
        return MemoryQueue(
            resource_name,
            maxsize=spec.get("maxsize", 0),
            batch_size=spec.get("batch_size", 100),
            poll_interval_s=spec.get("poll_interval_s", 0.1),
        )
    if resource_type == "interval":
        return IntervalSource.every(
            days=spec.get("days", 0),
            hours=spec.get("hours", 0),
            minutes=spec.get("minutes", 0),
            seconds=spec.get("seconds", 0),
            payload=spec.get("payload"),
            immediate=spec.get("immediate", False),
            overlap=spec.get("overlap", "allow"),
            timezone=spec.get("timezone"),
            timezone_name=spec.get("timezone_name"),
            poll_interval_s=spec.get("poll_interval_s"),
            name=resource_name,
        )
    if resource_type == "cron":
        return CronSource(
            _require_string(spec, "expression"),
            payload=spec.get("payload"),
            immediate=spec.get("immediate", False),
            overlap=spec.get("overlap", "allow"),
            timezone=spec.get("timezone"),
            timezone_name=spec.get("timezone_name"),
            poll_interval_s=spec.get("poll_interval_s", 1.0),
            name=resource_name,
        )
    if resource_type == "webhook":
        methods = spec.get("methods", ("POST",))
        return WebhookSource(
            path=_require_string(spec, "path"),
            methods=tuple(_string_list(methods, field=f"resources.{name}.methods")),
            host=spec.get("host", "127.0.0.1"),
            port=spec.get("port", 8080),
            parser=spec.get("parser", "auto"),
            auth=_build_webhook_auth(spec.get("auth")),
            response=_build_webhook_response(spec.get("response")),
            max_body_bytes=spec.get("max_body_bytes", 1024 * 1024),
            read_timeout_s=spec.get("read_timeout_s", 5.0),
            queue_maxsize=spec.get("queue_maxsize", 1000),
            batch_size=spec.get("batch_size", 100),
            poll_interval_s=spec.get("poll_interval_s", 0.1),
            name=resource_name,
        )
    if resource_type == "rabbitmq":
        return RabbitMQConnector(
            _require_string(spec, "url"),
            options=_mapping_value(spec.get("options"), field=f"resources.{name}.options"),
        )
    if resource_type == "rabbitmq_queue":
        connector = _resolve_dependency(resolve, spec, "connector")
        if not hasattr(connector, "queue"):
            raise TypeError(f"resource {spec['connector']!r} cannot build rabbitmq_queue")
        queue_name = _resource_name(spec, fallback=name, key="queue")
        return connector.queue(
            queue_name,
            routing_key=spec.get("routing_key"),
            exchange=spec.get("exchange"),
            exchange_type=spec.get("exchange_type", "direct"),
            bind=spec.get("bind", True),
            bind_arguments=_mapping_value(spec.get("bind_arguments"), field=f"resources.{name}.bind_arguments"),
            durable=spec.get("durable", True),
            auto_delete=spec.get("auto_delete", False),
            arguments=_mapping_value(spec.get("arguments"), field=f"resources.{name}.arguments"),
            exchange_durable=spec.get("exchange_durable"),
            exchange_auto_delete=spec.get("exchange_auto_delete", False),
            exchange_arguments=_mapping_value(
                spec.get("exchange_arguments"),
                field=f"resources.{name}.exchange_arguments",
            ),
            prefetch=spec.get("prefetch", 100),
            batch_size=spec.get("batch_size", 100),
            poll_interval_s=spec.get("poll_interval_s", 1.0),
            publisher_confirms=spec.get("publisher_confirms", True),
            persistent=spec.get("persistent", True),
        )
    if resource_type == "sqs":
        return SQSConnector(
            region_name=spec.get("region_name"),
            options=_mapping_value(spec.get("options"), field=f"resources.{name}.options"),
        )
    if resource_type == "sqs_queue":
        connector = _resolve_dependency(resolve, spec, "connector")
        if not hasattr(connector, "queue"):
            raise TypeError(f"resource {spec['connector']!r} cannot build sqs_queue")
        return connector.queue(
            _require_string(spec, "url"),
            wait_time_s=spec.get("wait_time_s", 20),
            visibility_timeout=spec.get("visibility_timeout"),
            batch_size=spec.get("batch_size", 10),
            poll_interval_s=spec.get("poll_interval_s", 0.0),
            message_group_id=spec.get("message_group_id"),
            deduplication_id_factory=_optional_ref(
                spec.get("deduplication_id_factory"),
                field=f"resources.{name}.deduplication_id_factory",
            ),
            on_fail=spec.get("on_fail", "leave"),
            delete_batch_size=spec.get("delete_batch_size", 10),
            delete_flush_interval_s=spec.get("delete_flush_interval_s", 0.5),
            heartbeat_interval_s=spec.get("heartbeat_interval_s"),
            heartbeat_visibility_timeout=spec.get("heartbeat_visibility_timeout"),
        )
    if resource_type == "mysql":
        return MySQLConnector(
            _require_string(spec, "dsn"),
            **_mapping_value(spec.get("engine_options"), field=f"resources.{name}.engine_options"),
        )
    if resource_type == "mysql_state_store":
        connector = _resolve_dependency(resolve, spec, "connector")
        if not hasattr(connector, "state_store"):
            raise TypeError(f"resource {spec['connector']!r} cannot build mysql_state_store")
        return connector.state_store(
            table=spec.get("table", "onestep_state"),
            key_column=spec.get("key_column", "state_key"),
            value_column=spec.get("value_column", "state_value"),
            updated_at_column=spec.get("updated_at_column", "updated_at"),
            auto_create=spec.get("auto_create", True),
        )
    if resource_type == "mysql_cursor_store":
        connector = _resolve_dependency(resolve, spec, "connector")
        if not hasattr(connector, "cursor_store"):
            raise TypeError(f"resource {spec['connector']!r} cannot build mysql_cursor_store")
        return connector.cursor_store(
            table=spec.get("table", "onestep_cursor"),
            key_column=spec.get("key_column", "cursor_key"),
            value_column=spec.get("value_column", "cursor_value"),
            updated_at_column=spec.get("updated_at_column", "updated_at"),
            auto_create=spec.get("auto_create", True),
        )
    if resource_type == "mysql_table_queue":
        connector = _resolve_dependency(resolve, spec, "connector")
        if not hasattr(connector, "table_queue"):
            raise TypeError(f"resource {spec['connector']!r} cannot build mysql_table_queue")
        return connector.table_queue(
            table=_require_string(spec, "table"),
            key=_require_string(spec, "key"),
            where=_require_string(spec, "where"),
            claim=_require_mapping(spec, "claim"),
            ack=_require_mapping(spec, "ack"),
            nack=_optional_mapping(spec.get("nack"), field=f"resources.{name}.nack") or None,
            batch_size=spec.get("batch_size", 100),
            poll_interval_s=spec.get("poll_interval_s", 1.0),
        )
    if resource_type == "mysql_incremental":
        connector = _resolve_dependency(resolve, spec, "connector")
        if not hasattr(connector, "incremental"):
            raise TypeError(f"resource {spec['connector']!r} cannot build mysql_incremental")
        raw_state_name = spec.get("state")
        state = None
        if raw_state_name is not None:
            state_name = _string_value(raw_state_name, field=f"resources.{name}.state")
            state = resolve(state_name)
            if not _is_cursor_store(state):
                raise TypeError(f"resource {state_name!r} cannot be used as incremental state")
        return connector.incremental(
            table=_require_string(spec, "table"),
            key=_require_string(spec, "key"),
            cursor=tuple(_string_list(spec.get("cursor"), field=f"resources.{name}.cursor")),
            where=spec.get("where"),
            batch_size=spec.get("batch_size", 1000),
            poll_interval_s=spec.get("poll_interval_s", 1.0),
            state=state,
            state_key=spec.get("state_key"),
        )
    if resource_type == "mysql_table_sink":
        connector = _resolve_dependency(resolve, spec, "connector")
        if not hasattr(connector, "table_sink"):
            raise TypeError(f"resource {spec['connector']!r} cannot build mysql_table_sink")
        keys = spec.get("keys")
        return connector.table_sink(
            table=_require_string(spec, "table"),
            mode=spec.get("mode", "insert"),
            keys=tuple(_string_list(keys, field=f"resources.{name}.keys")) if keys is not None else (),
        )

    raise ValueError(f"unsupported resource type {resource_type!r} for resource {name!r}")


def _build_handler(raw_handler: Any, *, task_name: str | None):
    if raw_handler is None:
        raise ValueError("task handler is required")
    if isinstance(raw_handler, str):
        ref = raw_handler
        params: Mapping[str, Any] = {}
    elif isinstance(raw_handler, Mapping):
        ref = _require_string(raw_handler, "ref")
        params = raw_handler.get("params", {})
        if not isinstance(params, Mapping):
            raise TypeError("'handler.params' must be a mapping")
    else:
        raise TypeError("'handler' must be a string ref or mapping")

    func = _load_ref(ref)
    if not callable(func):
        raise TypeError(f"handler ref {ref!r} did not resolve to a callable")
    if not params:
        return func

    bound_params = copy.deepcopy(dict(params))

    async def bound_handler(ctx, payload):
        result = func(ctx, payload, **copy.deepcopy(bound_params))
        if inspect.isawaitable(result):
            return await result
        return result

    bound_handler.__name__ = task_name or getattr(func, "__name__", "configured_handler")
    bound_handler.__doc__ = getattr(func, "__doc__", None)
    return bound_handler


def _build_retry(raw_retry: Any) -> RetryPolicy | None:
    if raw_retry is None:
        return None
    if isinstance(raw_retry, str):
        retry_type = raw_retry
        config: Mapping[str, Any] = {}
    elif isinstance(raw_retry, Mapping):
        retry_type = _require_string(raw_retry, "type")
        config = raw_retry
    else:
        raise TypeError("'retry' must be a string or mapping")

    normalized = retry_type.strip().lower().replace("-", "_")
    if normalized in {"no_retry", "none"}:
        return NoRetry()
    if normalized == "max_attempts":
        return MaxAttempts(
            max_attempts=config.get("max_attempts", 3),
            delay_s=config.get("delay_s"),
        )
    raise ValueError(f"unsupported retry type {retry_type!r}")


def _build_webhook_auth(raw_auth: Any) -> BearerAuth | None:
    if raw_auth is None:
        return None
    if isinstance(raw_auth, str):
        return BearerAuth(raw_auth)
    if not isinstance(raw_auth, Mapping):
        raise TypeError("'auth' must be a string or mapping")
    auth_type = str(raw_auth.get("type", "bearer")).strip().lower().replace("-", "_")
    if auth_type != "bearer":
        raise ValueError(f"unsupported webhook auth type {auth_type!r}")
    return BearerAuth(
        _require_string(raw_auth, "token"),
        header=raw_auth.get("header", "authorization"),
        scheme=raw_auth.get("scheme", "Bearer"),
        realm=raw_auth.get("realm", "webhook"),
    )


def _build_webhook_response(raw_response: Any) -> WebhookResponse | None:
    if raw_response is None:
        return None
    if not isinstance(raw_response, Mapping):
        raise TypeError("'response' must be a mapping")
    return WebhookResponse(
        status_code=raw_response.get("status_code", 202),
        body=raw_response.get("body"),
        headers=_mapping_value(raw_response.get("headers"), field="response.headers"),
    )


def _resolve_optional_source(resources: Mapping[str, Any], value: Any, *, task_index: int) -> Source | None:
    if value is None:
        return None
    name = _string_value(value, field=f"tasks[{task_index}].source")
    resolved = _resolve_resource(resources, name)
    if not isinstance(resolved, Source):
        raise TypeError(f"resource {name!r} cannot be used as a source")
    return resolved


def _resolve_optional_sinks(
    resources: Mapping[str, Any],
    value: Any,
    *,
    field: str,
    task_index: int,
) -> tuple[Sink, ...] | None:
    if value is None:
        return None
    names = _string_list(value, field=f"tasks[{task_index}].{field}")
    sinks: list[Sink] = []
    for name in names:
        resolved = _resolve_resource(resources, name)
        if not isinstance(resolved, Sink):
            raise TypeError(f"resource {name!r} cannot be used as a sink")
        sinks.append(resolved)
    return tuple(sinks)


def _resolve_app_state(resources: Mapping[str, Any], value: Any) -> Any:
    name = _string_value(value, field="app.state")
    resolved = _resolve_resource(resources, name)
    if not _is_state_store(resolved):
        raise TypeError(f"resource {name!r} cannot be used as app state")
    return resolved


def _resolve_resource(resources: Mapping[str, Any], name: str) -> Any:
    try:
        return resources[name]
    except KeyError as exc:
        raise KeyError(f"unknown resource {name!r}") from exc


def _resolve_dependency(resolve: Callable[[str], Any], spec: Mapping[str, Any], key: str) -> Any:
    return resolve(_require_string(spec, key))


def _load_ref(ref: str) -> Any:
    module_name, separator, attr_path = ref.partition(":")
    if not separator or not module_name or not attr_path:
        raise ValueError(f"invalid ref {ref!r}; expected 'package.module:attr'")
    module = importlib.import_module(module_name)
    value: Any = module
    for part in attr_path.split("."):
        value = getattr(value, part)
    return value


def _resource_name(spec: Mapping[str, Any], *, fallback: str, key: str = "name") -> str:
    value = spec.get(key)
    if value is None:
        return fallback
    return _string_value(value, field=key)


def _require_string(mapping: Mapping[str, Any], key: str) -> str:
    value = mapping.get(key)
    if not isinstance(value, str) or not value.strip():
        raise ValueError(f"'{key}' must be a non-empty string")
    return value


def _optional_string(mapping: Mapping[str, Any], key: str) -> str | None:
    value = mapping.get(key)
    if value is None:
        return None
    return _string_value(value, field=key)


def _string_value(value: Any, *, field: str) -> str:
    if not isinstance(value, str) or not value.strip():
        raise ValueError(f"'{field}' must be a non-empty string")
    return value


def _string_list(value: Any, *, field: str) -> list[str]:
    if value is None:
        raise ValueError(f"'{field}' must be provided")
    if isinstance(value, str):
        return [_string_value(value, field=field)]
    if not isinstance(value, Sequence) or isinstance(value, (str, bytes)):
        raise TypeError(f"'{field}' must be a string or list of strings")
    return [_string_value(item, field=field) for item in value]


def _require_mapping(mapping: Mapping[str, Any], key: str) -> dict[str, Any]:
    value = mapping.get(key)
    if not isinstance(value, Mapping):
        raise TypeError(f"'{key}' must be a mapping")
    return dict(value)


def _optional_mapping(value: Any, *, field: str) -> dict[str, Any] | None:
    if value is None:
        return None
    if not isinstance(value, Mapping):
        raise TypeError(f"'{field}' must be a mapping")
    return dict(value)


def _mapping_value(value: Any, *, field: str) -> dict[str, Any]:
    return _optional_mapping(value, field=field) or {}


def _optional_ref(value: Any, *, field: str) -> Any:
    if value is None:
        return None
    if callable(value):
        return value
    if isinstance(value, str):
        return _load_ref(value)
    raise TypeError(f"'{field}' must be a callable or ref string")


def _is_cursor_store(value: Any) -> bool:
    return hasattr(value, "load") and hasattr(value, "save")


def _is_state_store(value: Any) -> bool:
    return _is_cursor_store(value) and hasattr(value, "delete")


def _import_yaml():
    try:
        return importlib.import_module("yaml")
    except ModuleNotFoundError as exc:  # pragma: no cover - depends on environment
        raise ModuleNotFoundError(
            "YAML config support requires PyYAML. Install it with: pip install 'onestep[yaml]'"
        ) from exc



from __future__ import annotations

import argparse
import json
import os
import sys
from importlib.metadata import PackageNotFoundError, version

from .app import OneStepApp

_PROJECT_MARKERS = ("pyproject.toml", "setup.py", "setup.cfg")


def parse_args(argv: list[str] | None = None) -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="Run or inspect a OneStepApp target or YAML config")
    parser.add_argument(
        "--version",
        action="version",
        version=f"%(prog)s {_resolve_version()}",
    )
    subparsers = parser.add_subparsers(dest="command", required=True)

    run_parser = subparsers.add_parser("run", help="Load and run a OneStepApp target or YAML config")
    run_parser.add_argument("target", help="Python target (package.module:app) or path to *.yaml")

    check_parser = subparsers.add_parser("check", help="Load a target or YAML config and print its task summary")
    check_parser.add_argument("target", help="Python target (package.module:app) or path to *.yaml")
    check_parser.add_argument("--json", action="store_true", dest="as_json", help="Emit the summary as JSON")

    return parser.parse_args(_normalize_argv(argv))


def main(argv: list[str] | None = None) -> int:
    args = parse_args(argv)
    _ensure_local_import_paths()
    try:
        app = OneStepApp.load(args.target)
    except Exception as exc:
        print(f"onestep: failed to load {args.target}: {exc}", file=sys.stderr)
        return 2

    if args.command == "check":
        _print_summary(args.target, app, as_json=getattr(args, "as_json", False))
        return 0

    try:
        app.run()
    except Exception as exc:
        print(f"onestep: {args.target} failed while running: {exc}", file=sys.stderr)
        return 1
    return 0


def _ensure_local_import_paths() -> None:
    cwd = os.getcwd()
    if not cwd:
        return
    for path in reversed(_candidate_import_paths(cwd)):
        if _path_on_syspath(path):
            continue
        sys.path.insert(0, path)


def _candidate_import_paths(cwd: str) -> list[str]:
    candidates: list[str] = []
    seen: set[str] = set()

    def add(path: str) -> None:
        absolute_path = os.path.abspath(path)
        normalized_path = os.path.normcase(absolute_path)
        if normalized_path in seen or not os.path.isdir(absolute_path):
            return
        seen.add(normalized_path)
        candidates.append(absolute_path)

    add(cwd)
    add(os.path.join(cwd, "src"))

    project_root = _find_project_root(cwd)
    if project_root is not None:
        add(project_root)
        add(os.path.join(project_root, "src"))

    return candidates


def _find_project_root(start: str) -> str | None:
    current = os.path.abspath(start)
    while True:
        if any(os.path.exists(os.path.join(current, marker)) for marker in _PROJECT_MARKERS):
            return current
        parent = os.path.dirname(current)
        if parent == current:
            return None
        current = parent


def _path_on_syspath(path: str) -> bool:
    normalized_path = os.path.normcase(os.path.abspath(path))
    for entry in sys.path:
        current = entry or os.getcwd()
        if os.path.normcase(os.path.abspath(current)) == normalized_path:
            return True
    return False


def _normalize_argv(argv: list[str] | None) -> list[str] | None:
    if argv is None:
        argv = sys.argv[1:]
    if not argv:
        return argv
    if argv[0].startswith("-") or argv[0] in {"run", "check"}:
        return argv
    return ["run", *argv]


def _print_summary(target: str, app: OneStepApp, *, as_json: bool) -> None:
    summary = {
        "target": target,
        **app.describe(),
    }
    if as_json:
        print(json.dumps(summary, indent=2))
        return

    print(f"Target: {summary['target']}")
    print(f"App: {summary['name']}")
    print(f"Shutdown timeout: {_format_timeout(summary['shutdown_timeout_s'])}")
    print(f"Tasks: {len(summary['tasks'])}")
    for task in summary["tasks"]:
        source = _format_resource(task["source"])
        emit = _format_resources(task["emit"])
        dead_letter = _format_resources(task["dead_letter"])
        timeout = _format_timeout(task["timeout_s"])
        description = f" description={task['description']!r}" if task.get("description") else ""
        print(
            f"- {task['name']} source={source} emit={emit} dead_letter={dead_letter} "
            f"concurrency={task['concurrency']} timeout={timeout} retry={task['retry']}{description}"
        )


def _format_timeout(value: float | None) -> str:
    if value is None:
        return "-"
    return f"{value:.2f}s"


def _format_resources(items: list[dict[str, str]]) -> str:
    if not items:
        return "-"
    return ",".join(_format_resource(item) for item in items)


def _format_resource(item: dict[str, str] | None) -> str:
    if item is None:
        return "-"
    return f"{item['name']}<{item['type']}>"


def _resolve_version() -> str:
    try:
        return version("onestep")
    except PackageNotFoundError:
        return "dev"


if __name__ == "__main__":  # pragma: no cover
    sys.exit(main())



from __future__ import annotations

import asyncio
import contextlib
import importlib
import inspect
import logging
import signal
from collections.abc import Callable, Mapping, Sequence
from typing import Any

from .connectors.base import Sink, Source
from .events import TaskEvent
from .retry import RetryPolicy
from .runtime.runner import TaskRunner
from .state import InMemoryStateStore, StateStore
from .task import TaskHandler, TaskSpec


class OneStepApp:
    def __init__(
        self,
        name: str,
        *,
        config: Mapping[str, Any] | None = None,
        state: StateStore | None = None,
        shutdown_timeout_s: float | None = 30.0,
    ) -> None:
        if shutdown_timeout_s is not None and shutdown_timeout_s <= 0:
            raise ValueError("shutdown_timeout_s must be > 0")
        self.name = name
        self.config = dict(config or {})
        self.state = state or InMemoryStateStore()
        self.shutdown_timeout_s = shutdown_timeout_s
        self._tasks: list[TaskSpec] = []
        self._shutdown: asyncio.Event | None = None
        self._shutdown_requested = False
        self._resources: list[Any] = []
        self._startup_hooks: list[Callable[..., Any]] = []
        self._shutdown_hooks: list[Callable[..., Any]] = []
        self._event_handlers: list[Callable[..., Any]] = []
        self._loop: asyncio.AbstractEventLoop | None = None
        self._events_logger = logging.getLogger(f"onestep.{name}.events")

    @property
    def tasks(self) -> tuple[TaskSpec, ...]:
        return tuple(self._tasks)

    @property
    def is_stopping(self) -> bool:
        return self._shutdown_requested or (self._shutdown.is_set() if self._shutdown is not None else False)

    def request_shutdown(self) -> None:
        self._shutdown_requested = True
        try:
            current_loop = asyncio.get_running_loop()
        except RuntimeError:
            current_loop = None
        if self._shutdown is None:
            return
        if self._loop is None or self._loop is current_loop:
            self._shutdown.set()
            return
        self._loop.call_soon_threadsafe(self._shutdown.set)

    async def wait_for_shutdown(self) -> None:
        shutdown = self._ensure_shutdown_event()
        await shutdown.wait()

    def on_startup(self, func: Callable[..., Any] | None = None):
        return self._register_hook(self._startup_hooks, func)

    def on_shutdown(self, func: Callable[..., Any] | None = None):
        return self._register_hook(self._shutdown_hooks, func)

    def on_event(self, func: Callable[..., Any] | None = None):
        return self._register_hook(self._event_handlers, func)

    def task(
        self,
        *,
        name: str | None = None,
        description: str | None = None,
        source: Source | None = None,
        emit: Sink | Sequence[Sink] | None = None,
        dead_letter: Sink | Sequence[Sink] | None = None,
        concurrency: int = 1,
        retry: RetryPolicy | None = None,
        timeout_s: float | None = None,
    ):
        def decorator(func: TaskHandler) -> TaskHandler:
            task = TaskSpec.build(
                name=name or func.__name__,
                description=description,
                handler=func,
                source=source,
                sinks=emit,
                dead_letter=dead_letter,
                concurrency=concurrency,
                retry=retry,
                timeout_s=timeout_s,
            )
            self._tasks.append(task)
            return func

        return decorator

    async def startup(self) -> None:
        self._loop = asyncio.get_running_loop()
        self._shutdown_requested = False
        self._shutdown = asyncio.Event()
        resources: list[Any] = []
        seen: set[int] = set()
        for task in self._tasks:
            if task.source is not None and id(task.source) not in seen:
                resources.append(task.source)
                seen.add(id(task.source))
            for sink in task.sinks:
                if id(sink) not in seen:
                    resources.append(sink)
                    seen.add(id(sink))
            for sink in task.dead_letter_sinks:
                if id(sink) not in seen:
                    resources.append(sink)
                    seen.add(id(sink))
        opened: list[Any] = []
        self._resources = []
        try:
            for resource in resources:
                await resource.open()
                opened.append(resource)
            self._resources = list(opened)
            await self._run_hooks(self._startup_hooks)
        except Exception:
            await self._close_resources(opened, suppress_exceptions=True)
            self._resources = []
            raise

    async def shutdown(self) -> None:
        self._shutdown_requested = True
        if self._shutdown is not None:
            self._shutdown.set()
        hook_error: BaseException | None = None
        try:
            await self._run_hooks(self._shutdown_hooks)
        except BaseException as exc:
            hook_error = exc
        finally:
            close_error = await self._close_resources(self._resources, suppress_exceptions=False)
            self._resources = []
        if hook_error is not None:
            raise hook_error
        if close_error is not None:
            raise close_error

    async def serve(self) -> None:
        await self.startup()
        runners = [TaskRunner(self, task) for task in self._tasks if task.source is not None]
        try:
            if not runners:
                return
            await asyncio.gather(*(runner.run() for runner in runners))
        finally:
            await self.shutdown()

    def run(self) -> None:
        try:
            with self._install_signal_handlers():
                asyncio.run(self.serve())
        except KeyboardInterrupt:
            return None

    def describe(self) -> dict[str, Any]:
        return {
            "name": self.name,
            "shutdown_timeout_s": self.shutdown_timeout_s,
            "tasks": [
                {
                    "name": task.name,
                    "description": task.description,
                    "source": _describe_resource(task.source),
                    "emit": [_describe_resource(sink) for sink in task.sinks],
                    "dead_letter": [_describe_resource(sink) for sink in task.dead_letter_sinks],
                    "concurrency": task.concurrency,
                    "timeout_s": task.timeout_s,
                    "retry": task.retry.__class__.__name__,
                }
                for task in self._tasks
            ],
        }

    @classmethod
    def load(cls, target: str) -> "OneStepApp":
        from .config import is_yaml_target, load_yaml_app

        if is_yaml_target(target):
            return load_yaml_app(target)
        module_name, _, attr = target.partition(":")
        app_attr = attr or "app"
        module = importlib.import_module(module_name)
        value = getattr(module, app_attr)
        if isinstance(value, cls):
            return value
        if callable(value):
            resolved = _invoke_app_factory(target, value)
            if isinstance(resolved, cls):
                return resolved
        raise TypeError(f"{target} did not resolve to OneStepApp or a zero-argument factory")

    def _register_hook(
        self,
        storage: list[Callable[..., Any]],
        func: Callable[..., Any] | None,
    ):
        def decorator(callback: Callable[..., Any]) -> Callable[..., Any]:
            storage.append(callback)
            return callback

        if func is None:
            return decorator
        return decorator(func)

    async def _run_hooks(self, hooks: Sequence[Callable[..., Any]]) -> None:
        for hook in hooks:
            result = _invoke_hook(hook, self)
            if inspect.isawaitable(result):
                await result

    async def emit_event(self, event: TaskEvent) -> None:
        for handler in self._event_handlers:
            try:
                result = _invoke_single_arg(handler, event)
                if inspect.isawaitable(result):
                    await result
            except Exception:
                self._events_logger.exception("event handler failed", extra={"event_kind": event.kind.value})

    async def _close_resources(
        self,
        resources: Sequence[Any],
        *,
        suppress_exceptions: bool,
    ) -> BaseException | None:
        first_error: BaseException | None = None
        for resource in reversed(resources):
            try:
                await resource.close()
            except BaseException as exc:
                if first_error is None:
                    first_error = exc
                self._events_logger.exception(
                    "resource close failed",
                    extra={"resource_name": getattr(resource, "name", resource.__class__.__name__)},
                )
        if first_error is not None and not suppress_exceptions:
            return first_error
        return None

    def _ensure_shutdown_event(self) -> asyncio.Event:
        current_loop = asyncio.get_running_loop()
        if self._shutdown is None or self._loop is not current_loop:
            self._loop = current_loop
            self._shutdown = asyncio.Event()
            if self._shutdown_requested:
                self._shutdown.set()
        return self._shutdown

    @contextlib.contextmanager
    def _install_signal_handlers(self):
        installed: list[tuple[int, Any]] = []

        def handle_signal(signum, frame) -> None:
            self.request_shutdown()

        for sig_name in ("SIGINT", "SIGTERM"):
            sig = getattr(signal, sig_name, None)
            if sig is None:
                continue
            try:
                previous = signal.getsignal(sig)
                signal.signal(sig, handle_signal)
            except (ValueError, OSError):
                continue
            installed.append((sig, previous))
        try:
            yield
        finally:
            for sig, previous in reversed(installed):
                with contextlib.suppress(ValueError, OSError):
                    signal.signal(sig, previous)


def _invoke_hook(hook: Callable[..., Any], app: OneStepApp) -> Any:
    try:
        signature = inspect.signature(hook)
    except (TypeError, ValueError):
        return hook(app)

    positional = [
        parameter
        for parameter in signature.parameters.values()
        if parameter.kind in (inspect.Parameter.POSITIONAL_ONLY, inspect.Parameter.POSITIONAL_OR_KEYWORD)
    ]
    has_varargs = any(parameter.kind is inspect.Parameter.VAR_POSITIONAL for parameter in signature.parameters.values())
    if not positional and not has_varargs:
        return hook()
    return hook(app)


def _invoke_single_arg(callback: Callable[..., Any], value: Any) -> Any:
    try:
        signature = inspect.signature(callback)
    except (TypeError, ValueError):
        return callback(value)

    positional = [
        parameter
        for parameter in signature.parameters.values()
        if parameter.kind in (inspect.Parameter.POSITIONAL_ONLY, inspect.Parameter.POSITIONAL_OR_KEYWORD)
    ]
    has_varargs = any(parameter.kind is inspect.Parameter.VAR_POSITIONAL for parameter in signature.parameters.values())
    if not positional and not has_varargs:
        return callback()
    return callback(value)


def _describe_resource(resource: Source | Sink | None) -> dict[str, str] | None:
    if resource is None:
        return None
    return {
        "name": resource.name,
        "type": resource.__class__.__name__,
    }


def _invoke_app_factory(target: str, factory: Callable[..., Any]) -> Any:
    try:
        signature = inspect.signature(factory)
    except (TypeError, ValueError):
        return factory()

    required_positional = [
        parameter
        for parameter in signature.parameters.values()
        if parameter.kind in (inspect.Parameter.POSITIONAL_ONLY, inspect.Parameter.POSITIONAL_OR_KEYWORD)
        and parameter.default is inspect._empty
    ]
    required_keyword_only = [
        parameter
        for parameter in signature.parameters.values()
        if parameter.kind is inspect.Parameter.KEYWORD_ONLY and parameter.default is inspect._empty
    ]
    if required_positional or required_keyword_only:
        raise TypeError(f"{target} factory must not require arguments")
    return factory()



from __future__ import annotations

import inspect
from collections.abc import Callable, Sequence
from dataclasses import dataclass
from typing import Any

from .connectors.base import Sink, Source
from .retry import NoRetry, RetryPolicy

TaskHandler = Callable[["TaskContext", Any], Any]


@dataclass
class TaskSpec:
    name: str
    description: str | None
    handler: TaskHandler
    source: Source | None
    sinks: tuple[Sink, ...]
    dead_letter_sinks: tuple[Sink, ...]
    concurrency: int
    retry: RetryPolicy
    timeout_s: float | None

    @classmethod
    def build(
        cls,
        *,
        name: str,
        description: str | None,
        handler: TaskHandler,
        source: Source | None,
        sinks: Sink | Sequence[Sink] | None,
        dead_letter: Sink | Sequence[Sink] | None,
        concurrency: int,
        retry: RetryPolicy | None,
        timeout_s: float | None,
    ) -> "TaskSpec":
        if concurrency < 1:
            raise ValueError("concurrency must be >= 1")
        if timeout_s is not None and timeout_s <= 0:
            raise ValueError("timeout_s must be > 0")
        resolved_sinks = _normalize_sinks(sinks)
        resolved_dead_letter_sinks = _normalize_sinks(dead_letter)
        return cls(
            name=name,
            description=_normalize_description(description) or inspect.getdoc(handler),
            handler=handler,
            source=source,
            sinks=resolved_sinks,
            dead_letter_sinks=resolved_dead_letter_sinks,
            concurrency=concurrency,
            retry=retry or NoRetry(),
            timeout_s=timeout_s,
        )


def _normalize_sinks(sinks: Sink | Sequence[Sink] | None) -> tuple[Sink, ...]:
    if sinks is None:
        return ()
    if isinstance(sinks, Sequence) and not isinstance(sinks, (str, bytes)):
        return tuple(sinks)
    return (sinks,)


def _normalize_description(description: str | None) -> str | None:
    if description is None:
        return None
    normalized = description.strip()
    return normalized or None



from onestep import CronSource, OneStepApp

app = OneStepApp("cron-demo")


@app.task(source=CronSource("0 * * * *", timezone="Asia/Shanghai", overlap="skip"))
async def run_hourly(ctx, item):
    print("scheduled_at:", ctx.current.meta["scheduled_at"], "payload:", item)



import asyncio

from onestep import ConnectorOperation, ConnectorOperationError, SQSConnector


class FakeSQSClient:
    def __init__(self):
        self.available = []
        self.inflight = {}
        self.deleted = []
        self.deleted_batches = []
        self.visibility_changes = []
        self.sent = []
        self._counter = 0

    def send_message(self, **kwargs):
        self._counter += 1
        receipt_handle = f"rh-{self._counter}"
        message = {
            "MessageId": f"msg-{self._counter}",
            "ReceiptHandle": receipt_handle,
            "Body": kwargs["MessageBody"],
        }
        self.sent.append(kwargs)
        self.available.append(message)
        return {"MessageId": message["MessageId"]}

    def receive_message(self, **kwargs):
        take = kwargs["MaxNumberOfMessages"]
        messages = []
        while self.available and len(messages) < take:
            message = self.available.pop(0)
            self.inflight[message["ReceiptHandle"]] = message
            messages.append(message)
        if not messages:
            return {}
        return {"Messages": messages}

    def delete_message(self, QueueUrl, ReceiptHandle):
        self.deleted.append((QueueUrl, ReceiptHandle))
        self.inflight.pop(ReceiptHandle, None)

    def delete_message_batch(self, QueueUrl, Entries):
        self.deleted_batches.append((QueueUrl, Entries))
        for entry in Entries:
            receipt_handle = entry["ReceiptHandle"]
            self.deleted.append((QueueUrl, receipt_handle))
            self.inflight.pop(receipt_handle, None)
        return {"Successful": [{"Id": entry["Id"]} for entry in Entries], "Failed": []}

    def change_message_visibility(self, QueueUrl, ReceiptHandle, VisibilityTimeout):
        self.visibility_changes.append((QueueUrl, ReceiptHandle, VisibilityTimeout))
        message = self.inflight.get(ReceiptHandle)
        if message is not None and VisibilityTimeout == 0:
            self.available.append(message)
            self.inflight.pop(ReceiptHandle, None)


def test_sqs_queue_send_fetch_batch_delete_and_fail_delete():
    async def scenario():
        client = FakeSQSClient()
        connector = SQSConnector(region_name="ap-southeast-1", client=client)
        queue = connector.queue(
            "https://sqs.ap-southeast-1.amazonaws.com/123456789/jobs.fifo",
            wait_time_s=0,
            message_group_id="workers",
            on_fail="delete",
            delete_batch_size=2,
            delete_flush_interval_s=60,
        )

        await queue.publish({"value": 1}, meta={"source": "test"}, attempts=3)
        await queue.publish({"value": 2})
        assert client.sent[0]["MessageGroupId"] == "workers"

        batch = await queue.fetch(2)
        assert len(batch) == 2
        assert batch[0].payload == {"value": 1}
        assert batch[0].envelope.meta == {"source": "test"}
        assert batch[0].envelope.attempts == 3

        first_receipt = batch[0]._message["ReceiptHandle"]
        second_receipt = batch[1]._message["ReceiptHandle"]
        await batch[0].ack()
        assert client.deleted_batches == []

        await batch[1].ack()
        assert len(client.deleted_batches) == 1
        entries = client.deleted_batches[0][1]
        assert {entry["ReceiptHandle"] for entry in entries} == {first_receipt, second_receipt}

        await queue.publish({"value": 3})
        failed = await queue.fetch(1)
        assert len(failed) == 1
        fail_receipt = failed[0]._message["ReceiptHandle"]
        await failed[0].fail(RuntimeError("boom"))
        await queue.close()
        assert any(handle == fail_receipt for _, handle in client.deleted)
        await connector.close()

    asyncio.run(scenario())


def test_sqs_queue_retry_and_heartbeat_release():
    async def scenario():
        client = FakeSQSClient()
        connector = SQSConnector(region_name="ap-southeast-1", client=client)
        queue = connector.queue(
            "https://sqs.ap-southeast-1.amazonaws.com/123456789/jobs",
            wait_time_s=0,
            on_fail="release",
            heartbeat_interval_s=0.01,
            heartbeat_visibility_timeout=30,
            delete_flush_interval_s=0,
        )

        await queue.publish({"value": 1})
        batch = await queue.fetch(1)
        assert len(batch) == 1
        receipt = batch[0]._message["ReceiptHandle"]

        await batch[0].start_processing()
        await asyncio.sleep(0.03)
        assert any(item == (queue.url, receipt, 30) for item in client.visibility_changes)

        await batch[0].retry(delay_s=5)
        assert client.visibility_changes[-1] == (queue.url, receipt, 5)
        client.available.append(client.inflight.pop(receipt))

        redelivery = await queue.fetch(1)
        assert len(redelivery) == 1
        release_receipt = redelivery[0]._message["ReceiptHandle"]
        await redelivery[0].start_processing()
        await redelivery[0].fail(RuntimeError("release"))
        assert client.visibility_changes[-1] == (queue.url, release_receipt, 0)

        await queue.close()
        await connector.close()

    asyncio.run(scenario())


def test_sqs_queue_send_maps_retryable_errors():
    class BrokenClient(FakeSQSClient):
        def send_message(self, **kwargs):
            raise TimeoutError("timeout")

    async def scenario():
        client = BrokenClient()
        connector = SQSConnector(region_name="ap-southeast-1", client=client)
        queue = connector.queue(
            "https://sqs.ap-southeast-1.amazonaws.com/123456789/jobs",
            wait_time_s=0,
        )

        try:
            await queue.publish({"value": 1})
        except ConnectorOperationError as exc:
            assert exc.operation is ConnectorOperation.SEND
        else:
            raise AssertionError("expected ConnectorOperationError")

    asyncio.run(scenario())


def test_sqs_batch_delete_keeps_pending_entries_when_delete_raises() -> None:
    class BrokenDeleteClient(FakeSQSClient):
        def delete_message_batch(self, QueueUrl, Entries):
            raise TimeoutError("timeout")

    async def scenario():
        client = BrokenDeleteClient()
        connector = SQSConnector(region_name="ap-southeast-1", client=client)
        queue = connector.queue(
            "https://sqs.ap-southeast-1.amazonaws.com/123456789/jobs",
            wait_time_s=0,
            delete_batch_size=2,
            delete_flush_interval_s=60,
        )

        await queue.publish({"value": 1})
        await queue.publish({"value": 2})
        batch = await queue.fetch(2)
        first_receipt = batch[0]._message["ReceiptHandle"]
        second_receipt = batch[1]._message["ReceiptHandle"]

        await batch[0].ack()
        try:
            await batch[1].ack()
        except TimeoutError:
            pass
        else:
            raise AssertionError("expected batch delete failure")

        assert [entry["ReceiptHandle"] for entry in queue._pending_delete] == [
            first_receipt,
            second_receipt,
        ]
        assert first_receipt in client.inflight
        assert second_receipt in client.inflight

        await connector.close()

    asyncio.run(scenario())



import json
import logging
import os

from onestep import (
    InMemoryMetrics,
    MemoryQueue,
    NoRetry,
    OneStepApp,
    StructuredEventLogger,
    WebhookResponse,
    WebhookSource,
)


HOST = os.getenv("HOST", "127.0.0.1")
PORT = int(os.getenv("PORT", "8090"))


class EventJsonFormatter(logging.Formatter):
    def format(self, record: logging.LogRecord) -> str:
        payload = {
            "level": record.levelname,
            "logger": record.name,
            "message": record.getMessage(),
        }
        for key in (
            "event_kind",
            "app_name",
            "task_name",
            "source_name",
            "attempts",
            "duration_s",
            "failure_kind",
            "failure_exception_type",
            "failure_message",
            "task_event_meta",
        ):
            if hasattr(record, key):
                payload[key] = getattr(record, key)
        return json.dumps(payload, ensure_ascii=False)


def configure_logging() -> logging.Logger:
    logger = logging.getLogger("showcase.events")
    logger.handlers = []
    logger.setLevel(logging.DEBUG)
    logger.propagate = False
    handler = logging.StreamHandler()
    handler.setFormatter(EventJsonFormatter())
    logger.addHandler(handler)
    return logger


jobs = MemoryQueue("showcase.jobs", poll_interval_s=0.01)
dead_letter = MemoryQueue("showcase.dead", poll_interval_s=0.01)
metrics = InMemoryMetrics()
app = OneStepApp("runtime-showcase")
app.on_event(metrics)
app.on_event(StructuredEventLogger(logger=configure_logging()))


@app.on_startup
def print_instructions(app: OneStepApp) -> None:
    print(f"Webhook endpoint: http://{HOST}:{PORT}/demo/webhook")
    print("Try:")
    print(
        f"""curl -X POST http://{HOST}:{PORT}/demo/webhook \\
  -H 'Content-Type: application/json' \\
  -d '{{"action":"ok","value":21}}'"""
    )
    print(
        f"""curl -X POST http://{HOST}:{PORT}/demo/webhook \\
  -H 'Content-Type: application/json' \\
  -d '{{"action":"fail","value":21}}'"""
    )
    print(
        f"""curl -X POST http://{HOST}:{PORT}/demo/webhook \\
  -H 'Content-Type: application/json' \\
  -d '{{"action":"slow","value":21}}'"""
    )
    print("Press Ctrl-C to stop.")


@app.on_shutdown
def print_metrics() -> None:
    print("Metrics snapshot:")
    print(json.dumps(metrics.snapshot(), indent=2, ensure_ascii=False))


@app.task(
    source=WebhookSource(
        path="/demo/webhook",
        methods=("POST",),
        host=HOST,
        port=PORT,
        response=WebhookResponse(status_code=202, body={"accepted": True}),
    ),
    emit=jobs,
)
async def ingest_webhook(ctx, event):
    return event["body"]


@app.task(source=jobs, retry=NoRetry(), dead_letter=dead_letter, timeout_s=0.5)
async def process_job(ctx, item):
    action = item.get("action", "ok")
    if action == "slow":
        import asyncio

        await asyncio.sleep(2.0)
        return None
    if action == "fail":
        raise RuntimeError("requested failure")
    print(f"processed payload: {item}")
    return None


@app.task(source=dead_letter)
async def inspect_dead_letter(ctx, item):
    print("dead-letter received:")
    print(json.dumps(item, indent=2, ensure_ascii=False))


if __name__ == "__main__":
    app.run()



from onestep import IntervalSource, OneStepApp

app = OneStepApp("interval-demo")


@app.task(
    source=IntervalSource.every(
        seconds=5,
        immediate=True,
        overlap="skip",
        payload={"job": "refresh-cache"},
    )
)
async def refresh_cache(ctx, item):
    sequence = ctx.current.meta["sequence"]
    scheduled_at = ctx.current.meta["scheduled_at"]
    print("tick:", sequence, "scheduled_at:", scheduled_at, "payload:", item)
    if sequence >= 2:
        ctx.app.request_shutdown()



from __future__ import annotations

from collections import Counter
from dataclasses import dataclass, field
from datetime import datetime, timezone
from enum import Enum
import logging
from typing import Any

from .retry import FailureInfo, FailureKind


class TaskEventKind(str, Enum):
    FETCHED = "fetched"
    STARTED = "started"
    SUCCEEDED = "succeeded"
    RETRIED = "retried"
    FAILED = "failed"
    DEAD_LETTERED = "dead_lettered"
    CANCELLED = "cancelled"


@dataclass(frozen=True)
class TaskEvent:
    kind: TaskEventKind
    app: str
    task: str
    source: str | None
    attempts: int
    emitted_at: datetime = field(default_factory=lambda: datetime.now(timezone.utc))
    duration_s: float | None = None
    failure: FailureInfo | None = None
    meta: dict[str, Any] = field(default_factory=dict)


class InMemoryMetrics:
    def __init__(self) -> None:
        self._kind_counts: Counter[TaskEventKind] = Counter()
        self._task_counts: Counter[tuple[str, TaskEventKind]] = Counter()
        self._failure_counts: Counter[tuple[TaskEventKind, FailureKind]] = Counter()

    def __call__(self, event: TaskEvent) -> None:
        self._kind_counts[event.kind] += 1
        self._task_counts[(event.task, event.kind)] += 1
        if event.failure is not None:
            self._failure_counts[(event.kind, event.failure.kind)] += 1

    def count(
        self,
        kind: TaskEventKind,
        *,
        task: str | None = None,
        failure_kind: FailureKind | None = None,
    ) -> int:
        if task is not None:
            return self._task_counts[(task, kind)]
        if failure_kind is not None:
            return self._failure_counts[(kind, failure_kind)]
        return self._kind_counts[kind]

    def snapshot(self) -> dict[str, Any]:
        return {
            "kinds": {kind.value: count for kind, count in self._kind_counts.items()},
            "tasks": {
                f"{task}:{kind.value}": count
                for (task, kind), count in self._task_counts.items()
            },
            "failures": {
                f"{kind.value}:{failure_kind.value}": count
                for (kind, failure_kind), count in self._failure_counts.items()
            },
        }


class StructuredEventLogger:
    def __init__(
        self,
        *,
        logger: logging.Logger | None = None,
        logger_name: str = "onestep.events",
        level_by_kind: dict[TaskEventKind, int] | None = None,
    ) -> None:
        self.logger = logger or logging.getLogger(logger_name)
        self.level_by_kind = {
            TaskEventKind.FETCHED: logging.DEBUG,
            TaskEventKind.STARTED: logging.DEBUG,
            TaskEventKind.SUCCEEDED: logging.INFO,
            TaskEventKind.RETRIED: logging.WARNING,
            TaskEventKind.FAILED: logging.ERROR,
            TaskEventKind.DEAD_LETTERED: logging.ERROR,
            TaskEventKind.CANCELLED: logging.WARNING,
            **(level_by_kind or {}),
        }

    def __call__(self, event: TaskEvent) -> None:
        failure = event.failure
        self.logger.log(
            self.level_by_kind[event.kind],
            f"task {event.kind.value}",
            extra={
                "event_kind": event.kind.value,
                "app_name": event.app,
                "task_name": event.task,
                "source_name": event.source,
                "attempts": event.attempts,
                "duration_s": event.duration_s,
                "emitted_at": event.emitted_at.isoformat(),
                "failure_kind": failure.kind.value if failure is not None else None,
                "failure_exception_type": failure.exception_type if failure is not None else None,
                "failure_message": failure.message if failure is not None else None,
                "task_event_meta": dict(event.meta),
            },
        )


__all__ = [
    "InMemoryMetrics",
    "StructuredEventLogger",
    "TaskEvent",
    "TaskEventKind",
]



import asyncio
from pathlib import Path

import sqlalchemy as sa

from onestep import MySQLConnector


def test_mysql_incremental_cursor_advances_in_order(tmp_path: Path) -> None:
    db_url = f"sqlite:///{tmp_path / 'incremental.db'}"
    engine = sa.create_engine(db_url, future=True)
    metadata = sa.MetaData()
    users = sa.Table(
        "users",
        metadata,
        sa.Column("id", sa.Integer, primary_key=True),
        sa.Column("name", sa.String, nullable=False),
        sa.Column("updated_at", sa.Integer, nullable=False),
        sa.Column("deleted", sa.Integer, nullable=False, default=0),
    )
    metadata.create_all(engine)
    with engine.begin() as conn:
        conn.execute(
            sa.insert(users),
            [
                {"id": 1, "name": "A", "updated_at": 10, "deleted": 0},
                {"id": 2, "name": "B", "updated_at": 10, "deleted": 0},
            ],
        )
    engine.dispose()

    async def scenario() -> None:
        db = MySQLConnector(db_url)
        state = db.cursor_store(table="onestep_cursor")
        source = db.incremental(
            table="users",
            key="id",
            cursor=("updated_at", "id"),
            where="deleted = 0",
            batch_size=10,
            poll_interval_s=0.01,
            state=state,
            state_key="users-sync",
        )

        batch = await source.fetch(10)
        assert [item.payload["id"] for item in batch] == [1, 2]

        await batch[1].ack()
        assert await state.load("users-sync") is None

        await batch[0].ack()
        assert await state.load("users-sync") == [10, 2]
        await db.close()

        restarted_db = MySQLConnector(db_url)
        restarted_state = restarted_db.cursor_store(table="onestep_cursor")
        restarted_source = restarted_db.incremental(
            table="users",
            key="id",
            cursor=("updated_at", "id"),
            where="deleted = 0",
            batch_size=10,
            poll_interval_s=0.01,
            state=restarted_state,
            state_key="users-sync",
        )

        empty_batch = await restarted_source.fetch(10)
        assert empty_batch == []

        with restarted_db.engine.begin() as conn:
            conn.execute(sa.insert(users), [{"id": 3, "name": "C", "updated_at": 11, "deleted": 0}])

        next_batch = await restarted_source.fetch(10)
        assert [item.payload["id"] for item in next_batch] == [3]
        await next_batch[0].ack()
        assert await restarted_state.load("users-sync") == [11, 3]
        await restarted_db.close()

    asyncio.run(scenario())


def test_mysql_incremental_does_not_refetch_pending_gap_with_out_of_order_ack(tmp_path: Path) -> None:
    db_url = f"sqlite:///{tmp_path / 'incremental_gap.db'}"
    engine = sa.create_engine(db_url, future=True)
    metadata = sa.MetaData()
    users = sa.Table(
        "users",
        metadata,
        sa.Column("id", sa.Integer, primary_key=True),
        sa.Column("name", sa.String, nullable=False),
        sa.Column("updated_at", sa.Integer, nullable=False),
        sa.Column("deleted", sa.Integer, nullable=False, default=0),
    )
    metadata.create_all(engine)
    with engine.begin() as conn:
        conn.execute(
            sa.insert(users),
            [
                {"id": 1, "name": "A", "updated_at": 10, "deleted": 0},
                {"id": 2, "name": "B", "updated_at": 10, "deleted": 0},
            ],
        )
    engine.dispose()

    async def scenario() -> None:
        db = MySQLConnector(db_url)
        source = db.incremental(
            table="users",
            key="id",
            cursor=("updated_at", "id"),
            where="deleted = 0",
            batch_size=10,
            poll_interval_s=0.01,
        )

        batch = await source.fetch(10)
        assert [item.payload["id"] for item in batch] == [1, 2]

        await batch[1].ack()
        next_batch = await source.fetch(1)
        assert next_batch == []

        await batch[0].ack()
        await db.close()

    asyncio.run(scenario())


def test_mysql_incremental_uses_key_as_tie_breaker_when_cursor_is_not_unique(tmp_path: Path) -> None:
    db_url = f"sqlite:///{tmp_path / 'incremental_tiebreak.db'}"
    engine = sa.create_engine(db_url, future=True)
    metadata = sa.MetaData()
    users = sa.Table(
        "users",
        metadata,
        sa.Column("id", sa.Integer, primary_key=True),
        sa.Column("name", sa.String, nullable=False),
        sa.Column("updated_at", sa.Integer, nullable=False),
        sa.Column("deleted", sa.Integer, nullable=False, default=0),
    )
    metadata.create_all(engine)
    with engine.begin() as conn:
        conn.execute(
            sa.insert(users),
            [
                {"id": 1, "name": "A", "updated_at": 10, "deleted": 0},
                {"id": 2, "name": "B", "updated_at": 10, "deleted": 0},
            ],
        )
    engine.dispose()

    async def scenario() -> None:
        db = MySQLConnector(db_url)
        state = db.cursor_store(table="onestep_cursor")
        source = db.incremental(
            table="users",
            key="id",
            cursor=("updated_at",),
            where="deleted = 0",
            batch_size=10,
            poll_interval_s=0.01,
            state=state,
            state_key="users-updated-at",
        )

        batch = await source.fetch(10)
        assert [item.payload["id"] for item in batch] == [1, 2]
        assert source.cursor == ("updated_at", "id")

        await batch[0].ack()
        await batch[1].ack()
        assert await state.load("users-updated-at") == [10, 2]

        with db.engine.begin() as conn:
            conn.execute(sa.insert(users), [{"id": 3, "name": "C", "updated_at": 10, "deleted": 0}])

        next_batch = await source.fetch(10)
        assert [item.payload["id"] for item in next_batch] == [3]
        await next_batch[0].ack()
        await db.close()

    asyncio.run(scenario())


def test_mysql_incremental_default_state_key_separates_distinct_where_clauses(tmp_path: Path) -> None:
    db_url = f"sqlite:///{tmp_path / 'incremental_where_scope.db'}"
    engine = sa.create_engine(db_url, future=True)
    metadata = sa.MetaData()
    users = sa.Table(
        "users",
        metadata,
        sa.Column("id", sa.Integer, primary_key=True),
        sa.Column("name", sa.String, nullable=False),
        sa.Column("updated_at", sa.Integer, nullable=False),
        sa.Column("deleted", sa.Integer, nullable=False, default=0),
    )
    metadata.create_all(engine)
    with engine.begin() as conn:
        conn.execute(
            sa.insert(users),
            [
                {"id": 1, "name": "A", "updated_at": 10, "deleted": 0},
                {"id": 2, "name": "B", "updated_at": 11, "deleted": 0},
                {"id": 3, "name": "C", "updated_at": 9, "deleted": 1},
            ],
        )
    engine.dispose()

    async def scenario() -> None:
        db = MySQLConnector(db_url)
        state = db.cursor_store(table="onestep_cursor")
        active = db.incremental(
            table="users",
            key="id",
            cursor=("updated_at",),
            where="deleted = 0",
            batch_size=10,
            poll_interval_s=0.01,
            state=state,
        )
        deleted = db.incremental(
            table="users",
            key="id",
            cursor=("updated_at",),
            where="deleted = 1",
            batch_size=10,
            poll_interval_s=0.01,
            state=state,
        )

        assert active.state_key != deleted.state_key

        active_batch = await active.fetch(10)
        assert [item.payload["id"] for item in active_batch] == [1, 2]
        await active_batch[0].ack()
        await active_batch[1].ack()

        deleted_batch = await deleted.fetch(10)
        assert [item.payload["id"] for item in deleted_batch] == [3]
        await deleted_batch[0].ack()

        await db.close()

    asyncio.run(scenario())



from onestep import RabbitMQConnector, OneStepApp

app = OneStepApp("rabbitmq-demo")
rmq = RabbitMQConnector("amqp://guest:guest@localhost/")
queue = rmq.queue(
    "incoming_jobs",
    exchange="jobs.events",
    routing_key="jobs.created",
    prefetch=50,
)
processed = rmq.queue(
    "processed_jobs",
    exchange="jobs.events",
    routing_key="jobs.done",
)


@app.task(source=queue, emit=processed, concurrency=8)
async def process_job(ctx, item):
    return {"job": item["job"], "status": "done"}



from __future__ import annotations

import asyncio
import logging
import copy
import inspect
import time
from typing import TYPE_CHECKING

from onestep.context import TaskContext
from onestep.envelope import Envelope
from onestep.events import TaskEvent, TaskEventKind
from onestep.resilience import ConnectorOperationError, connector_retry_delay, is_retryable_connector_error
from onestep.retry import FailureInfo, FailureKind, RetryDecision, resolve_retry_action
from onestep.task import TaskSpec

if TYPE_CHECKING:
    from onestep.app import OneStepApp
    from onestep.connectors.base import Delivery


class TaskRunner:
    _SEND_ATTEMPTS = 2
    def __init__(self, app: "OneStepApp", task: TaskSpec) -> None:
        self.app = app
        self.task = task
        self._inflight: set[asyncio.Task[None]] = set()
        self._logger = logging.getLogger(f"onestep.{app.name}.{task.name}")

    async def run(self) -> None:
        if self.task.source is None:
            return
        try:
            while not self.app.is_stopping:
                available = self.task.concurrency - len(self._inflight)
                if available <= 0:
                    await self._wait_for_inflight(timeout=self.task.source.poll_interval_s)
                    continue

                deliveries = await self._fetch_deliveries(available)
                if not deliveries:
                    if self.app.is_stopping:
                        break
                    if self._inflight:
                        await self._wait_for_inflight(timeout=self.task.source.poll_interval_s)
                    else:
                        await asyncio.sleep(self.task.source.poll_interval_s)
                    continue
                await self._emit_batch_event(TaskEventKind.FETCHED, deliveries)

                for delivery in deliveries:
                    pending = asyncio.create_task(self._handle_delivery(delivery))
                    self._inflight.add(pending)
                    pending.add_done_callback(self._inflight.discard)
        finally:
            await self._drain_inflight()

    async def _wait_for_inflight(self, timeout: float | None) -> None:
        if not self._inflight:
            return
        await asyncio.wait(self._inflight, timeout=timeout, return_when=asyncio.FIRST_COMPLETED)

    async def _fetch_deliveries(self, limit: int) -> list["Delivery"]:
        if self.task.source is None or self.app.is_stopping:
            return []
        fetch_task = asyncio.create_task(self.task.source.fetch(limit))
        shutdown_task = asyncio.create_task(self.app.wait_for_shutdown())
        done, pending = await asyncio.wait(
            {fetch_task, shutdown_task},
            return_when=asyncio.FIRST_COMPLETED,
        )
        try:
            if fetch_task in done:
                try:
                    return fetch_task.result()
                except ConnectorOperationError as exc:
                    if not is_retryable_connector_error(exc):
                        raise
                    await self._handle_source_fetch_error(exc)
                    return []
            fetch_task.cancel()
            await asyncio.gather(fetch_task, return_exceptions=True)
            return []
        finally:
            for pending_task in pending:
                pending_task.cancel()
            await asyncio.gather(*pending, return_exceptions=True)

    async def _handle_source_fetch_error(self, exc: ConnectorOperationError) -> None:
        fallback_s = self.task.source.poll_interval_s if self.task.source is not None else 1.0
        delay_s = connector_retry_delay(exc, fallback_s=fallback_s)
        self._logger.warning(
            "source fetch degraded; backing off",
            extra={
                "connector_backend": exc.backend,
                "connector_operation": exc.operation.value,
                "connector_kind": exc.kind.value,
                "connector_retry_delay_s": delay_s,
            },
            exc_info=exc,
        )
        if delay_s <= 0 or self.app.is_stopping:
            return
        try:
            await asyncio.wait_for(self.app.wait_for_shutdown(), timeout=delay_s)
        except asyncio.TimeoutError:
            return

    async def _send_to_sink(self, sink, envelope: Envelope) -> None:
        fallback_s = getattr(sink, "poll_interval_s", 1.0)
        for attempt in range(self._SEND_ATTEMPTS):
            try:
                await sink.send(envelope)
                return
            except ConnectorOperationError as exc:
                if not is_retryable_connector_error(exc) or attempt == self._SEND_ATTEMPTS - 1:
                    raise
                delay_s = connector_retry_delay(exc, fallback_s=fallback_s)
                self._logger.warning(
                    "sink send degraded; retrying",
                    extra={
                        "connector_backend": exc.backend,
                        "connector_operation": exc.operation.value,
                        "connector_kind": exc.kind.value,
                        "connector_retry_delay_s": delay_s,
                    },
                    exc_info=exc,
                )
                if delay_s <= 0 or self.app.is_stopping:
                    continue
                try:
                    await asyncio.wait_for(self.app.wait_for_shutdown(), timeout=delay_s)
                except asyncio.TimeoutError:
                    continue
                raise

    async def _handle_delivery(self, delivery: "Delivery") -> None:
        ctx = TaskContext(app=self.app, task=self.task, delivery=delivery)
        started_at = time.perf_counter()
        try:
            await delivery.start_processing()
            await self._emit_event(TaskEventKind.STARTED, delivery)
            result = await self._invoke_handler(ctx, delivery)
            if result is not None and self.task.sinks:
                envelope = Envelope(body=result)
                for sink in self.task.sinks:
                    await self._send_to_sink(sink, envelope)
            await delivery.ack()
            await self._emit_event(
                TaskEventKind.SUCCEEDED,
                delivery,
                duration_s=time.perf_counter() - started_at,
            )
        except asyncio.CancelledError:
            failure = FailureInfo.from_exception(asyncio.CancelledError(), kind=FailureKind.CANCELLED)
            ctx.logger.warning("task cancelled", extra={"failure_kind": failure.kind.value})
            duration_s = time.perf_counter() - started_at
            await self._emit_event(
                TaskEventKind.CANCELLED,
                delivery,
                failure=failure,
                duration_s=duration_s,
            )
            await delivery.retry()
            await self._emit_event(
                TaskEventKind.RETRIED,
                delivery,
                failure=failure,
                duration_s=duration_s,
            )
            raise
        except asyncio.TimeoutError as exc:
            await self._handle_failure(
                ctx,
                delivery,
                exc,
                FailureKind.TIMEOUT,
                duration_s=time.perf_counter() - started_at,
            )
        except Exception as exc:
            await self._handle_failure(
                ctx,
                delivery,
                exc,
                FailureKind.ERROR,
                duration_s=time.perf_counter() - started_at,
            )

    async def _invoke_handler(self, ctx: TaskContext, delivery: "Delivery"):
        result = self.task.handler(ctx, delivery.payload)
        if inspect.isawaitable(result):
            if self.task.timeout_s is not None:
                return await asyncio.wait_for(result, timeout=self.task.timeout_s)
            return await result
        return result

    async def _handle_failure(
        self,
        ctx: TaskContext,
        delivery: "Delivery",
        exc: Exception,
        kind: FailureKind,
        *,
        duration_s: float | None,
    ) -> None:
        failure = FailureInfo.from_exception(exc, kind=kind)
        ctx.logger.exception("task failed", extra={"failure_kind": failure.kind.value})
        action = resolve_retry_action(self.task.retry, delivery.envelope, exc, failure)
        if action.decision is RetryDecision.RETRY:
            await delivery.retry(delay_s=action.delay_s)
            await self._emit_event(
                TaskEventKind.RETRIED,
                delivery,
                failure=failure,
                duration_s=duration_s,
            )
            return
        if self.task.dead_letter_sinks:
            published = await self._publish_dead_letter(
                ctx,
                delivery,
                failure,
                duration_s=duration_s,
            )
            if not published:
                return
        await self._fail_delivery(ctx, delivery, exc)
        await self._emit_event(
            TaskEventKind.FAILED,
            delivery,
            failure=failure,
            duration_s=duration_s,
        )

    async def _publish_dead_letter(
        self,
        ctx: TaskContext,
        delivery: "Delivery",
        failure: FailureInfo,
        *,
        duration_s: float | None,
    ) -> bool:
        envelope = Envelope(
            body={
                "payload": copy.deepcopy(delivery.envelope.body),
                "failure": failure.as_dict(),
            },
            meta={
                "app": self.app.name,
                "task": self.task.name,
                "source": self.task.source.name if self.task.source is not None else None,
                "original_meta": copy.deepcopy(delivery.envelope.meta),
                "original_attempts": delivery.envelope.attempts,
            },
            attempts=0,
        )
        try:
            for sink in self.task.dead_letter_sinks:
                await sink.send(envelope)
        except Exception:
            ctx.logger.exception("dead-letter publish failed; retrying original delivery")
            await delivery.retry()
            await self._emit_event(
                TaskEventKind.RETRIED,
                delivery,
                failure=failure,
                duration_s=duration_s,
            )
            return False
        await self._emit_event(
            TaskEventKind.DEAD_LETTERED,
            delivery,
            failure=failure,
            duration_s=duration_s,
        )
        return True

    async def _fail_delivery(self, ctx: TaskContext, delivery: "Delivery", exc: Exception) -> None:
        try:
            await delivery.fail(exc)
        except Exception:
            ctx.logger.exception("delivery fail action failed; retrying original delivery")
            await delivery.retry()

    async def _drain_inflight(self) -> None:
        if not self._inflight:
            return
        if self.app.shutdown_timeout_s is None:
            await asyncio.gather(*self._inflight, return_exceptions=True)
            return
        done, pending = await asyncio.wait(self._inflight, timeout=self.app.shutdown_timeout_s)
        if done:
            await asyncio.gather(*done, return_exceptions=True)
        if pending:
            for pending_task in pending:
                pending_task.cancel()
            await asyncio.gather(*pending, return_exceptions=True)

    async def _emit_batch_event(self, kind: TaskEventKind, deliveries: list["Delivery"]) -> None:
        for delivery in deliveries:
            await self._emit_event(kind, delivery)

    async def _emit_event(
        self,
        kind: TaskEventKind,
        delivery: "Delivery",
        *,
        duration_s: float | None = None,
        failure: FailureInfo | None = None,
    ) -> None:
        event = TaskEvent(
            kind=kind,
            app=self.app.name,
            task=self.task.name,
            source=self.task.source.name if self.task.source is not None else None,
            attempts=delivery.envelope.attempts,
            duration_s=duration_s,
            failure=failure,
            meta=copy.deepcopy(delivery.envelope.meta),
        )
        await self.app.emit_event(event)



from __future__ import annotations

from enum import Enum
from typing import Any

try:  # pragma: no cover - optional dependency
    import aio_pika.exceptions as aio_pika_exceptions
except ImportError:  # pragma: no cover - optional dependency
    aio_pika_exceptions = None

try:  # pragma: no cover - optional dependency
    import aiormq.exceptions as aiormq_exceptions
except ImportError:  # pragma: no cover - optional dependency
    aiormq_exceptions = None

try:  # pragma: no cover - optional dependency
    import botocore.exceptions as botocore_exceptions
except ImportError:  # pragma: no cover - optional dependency
    botocore_exceptions = None

try:  # pragma: no cover - optional dependency
    import sqlalchemy as sa
except ImportError:  # pragma: no cover - optional dependency
    sa = None


class ConnectorErrorKind(str, Enum):
    DISCONNECTED = "disconnected"
    TRANSIENT = "transient"
    THROTTLED = "throttled"
    MISCONFIGURED = "misconfigured"
    PERMANENT = "permanent"
    UNCERTAIN = "uncertain"


class ConnectorOperation(str, Enum):
    OPEN = "open"
    FETCH = "fetch"
    SEND = "send"
    ACK = "ack"
    RETRY = "retry"
    FAIL = "fail"
    CLOSE = "close"


class ConnectorOperationError(Exception):
    def __init__(
        self,
        *,
        backend: str,
        operation: ConnectorOperation,
        kind: ConnectorErrorKind,
        source_name: str | None = None,
        retry_delay_s: float | None = None,
        cause: BaseException | None = None,
        message: str | None = None,
    ) -> None:
        self.backend = backend
        self.operation = operation
        self.kind = kind
        self.source_name = source_name
        self.retry_delay_s = retry_delay_s
        self.cause = cause
        super().__init__(message or self._build_message())

    def _build_message(self) -> str:
        target = f" for {self.source_name}" if self.source_name else ""
        return f"{self.backend} {self.operation.value} failed ({self.kind.value}){target}"


def is_retryable_connector_error(exc: ConnectorOperationError | ConnectorErrorKind) -> bool:
    kind = exc.kind if isinstance(exc, ConnectorOperationError) else exc
    return kind in {
        ConnectorErrorKind.DISCONNECTED,
        ConnectorErrorKind.TRANSIENT,
        ConnectorErrorKind.THROTTLED,
    }


def connector_retry_delay(exc: ConnectorOperationError, *, fallback_s: float) -> float:
    if exc.retry_delay_s is not None:
        return max(0.0, exc.retry_delay_s)
    if exc.kind is ConnectorErrorKind.THROTTLED:
        return max(fallback_s, 1.0)
    return max(fallback_s, 0.0)


def classify_rabbitmq_error(exc: BaseException) -> ConnectorErrorKind | None:
    if isinstance(exc, (ConnectionError, OSError)):
        return ConnectorErrorKind.DISCONNECTED
    if isinstance(exc, RuntimeError):
        message = str(exc).lower()
        if "closed" in message or "no active transport" in message:
            return ConnectorErrorKind.DISCONNECTED

    if aio_pika_exceptions is not None:
        result = _classify_amqp_error(exc, aio_pika_exceptions)
        if result is not None:
            return result
    if aiormq_exceptions is not None:
        result = _classify_amqp_error(exc, aiormq_exceptions)
        if result is not None:
            return result
    return None


def classify_sqs_error(exc: BaseException) -> ConnectorErrorKind | None:
    if isinstance(exc, (ConnectionError, OSError, TimeoutError)):
        return ConnectorErrorKind.DISCONNECTED
    if botocore_exceptions is None:
        return None

    transient_types = []
    for name in ("EndpointConnectionError", "ConnectionClosedError", "ReadTimeoutError", "ConnectTimeoutError"):
        error_type = getattr(botocore_exceptions, name, None)
        if isinstance(error_type, type):
            transient_types.append(error_type)
    if transient_types and isinstance(exc, tuple(transient_types)):
        return ConnectorErrorKind.DISCONNECTED

    client_error = getattr(botocore_exceptions, "ClientError", None)
    if isinstance(client_error, type) and isinstance(exc, client_error):
        code = str(exc.response.get("Error", {}).get("Code", "")).lower()
        if code in {
            "throttling",
            "throttlingexception",
            "requestthrottled",
            "toomanyrequestsexception",
            "slowdown",
        }:
            return ConnectorErrorKind.THROTTLED
        if code in {"requesttimeout", "internalerror", "serviceunavailable"}:
            return ConnectorErrorKind.TRANSIENT
        if code in {
            "accessdenied",
            "accessdeniedexception",
            "invalidclienttokenid",
            "signaturedoesnotmatch",
            "aws.simplequeueservice.nonexistentqueue",
        }:
            return ConnectorErrorKind.MISCONFIGURED
        return ConnectorErrorKind.PERMANENT
    return None


def classify_sqlalchemy_error(exc: BaseException) -> ConnectorErrorKind | None:
    if sa is None:
        return None
    sql_exc = sa.exc
    if isinstance(exc, getattr(sql_exc, "TimeoutError", ())):
        return ConnectorErrorKind.TRANSIENT
    if isinstance(exc, getattr(sql_exc, "InterfaceError", ())):
        return ConnectorErrorKind.DISCONNECTED
    if isinstance(exc, getattr(sql_exc, "ProgrammingError", ())):
        return ConnectorErrorKind.PERMANENT
    if isinstance(exc, getattr(sql_exc, "DBAPIError", ())):
        if getattr(exc, "connection_invalidated", False):
            return ConnectorErrorKind.DISCONNECTED
        message = " ".join(
            str(part).lower()
            for part in (
                getattr(exc, "orig", None),
                exc,
            )
            if part is not None
        )
        if any(token in message for token in ("server has gone away", "lost connection", "connection refused")):
            return ConnectorErrorKind.DISCONNECTED
        if any(token in message for token in ("lock wait timeout", "deadlock found")):
            return ConnectorErrorKind.TRANSIENT
        if any(token in message for token in ("access denied", "unknown database", "authentication")):
            return ConnectorErrorKind.MISCONFIGURED
        if any(token in message for token in ("no such table", "unknown table", "unknown column", "syntax error")):
            return ConnectorErrorKind.PERMANENT
        if isinstance(exc, getattr(sql_exc, "OperationalError", ())):
            return ConnectorErrorKind.TRANSIENT
    return None


def as_connector_operation_error(
    *,
    backend: str,
    operation: ConnectorOperation,
    exc: BaseException,
    source_name: str | None = None,
    retry_delay_s: float | None = None,
) -> ConnectorOperationError | None:
    classifier = {
        "rabbitmq": classify_rabbitmq_error,
        "sqs": classify_sqs_error,
        "mysql": classify_sqlalchemy_error,
    }.get(backend)
    if classifier is None:
        return None
    kind = classifier(exc)
    if kind is None:
        return None
    return ConnectorOperationError(
        backend=backend,
        operation=operation,
        kind=kind,
        source_name=source_name,
        retry_delay_s=retry_delay_s,
        cause=exc,
    )


def _classify_amqp_error(exc: BaseException, module: Any) -> ConnectorErrorKind | None:
    disconnected_names = {
        "AMQPConnectionError",
        "AMQPChannelError",
        "ConnectionClosed",
        "ChannelClosed",
        "ChannelInvalidStateError",
        "DeliveryError",
    }
    misconfigured_names = {
        "AuthenticationError",
        "ProbableAuthenticationError",
        "IncompatibleProtocolError",
    }
    permanent_names = {
        "ChannelNotFoundEntity",
        "ChannelPreconditionFailed",
    }
    for name in disconnected_names:
        error_type = getattr(module, name, None)
        if isinstance(error_type, type) and isinstance(exc, error_type):
            return ConnectorErrorKind.DISCONNECTED
    for name in misconfigured_names:
        error_type = getattr(module, name, None)
        if isinstance(error_type, type) and isinstance(exc, error_type):
            return ConnectorErrorKind.MISCONFIGURED
    for name in permanent_names:
        error_type = getattr(module, name, None)
        if isinstance(error_type, type) and isinstance(exc, error_type):
            return ConnectorErrorKind.PERMANENT
    return None



import asyncio
from datetime import datetime, timedelta
from zoneinfo import ZoneInfo

from onestep import CronSource, IntervalSource


class FakeClock:
    def __init__(self, current: datetime) -> None:
        self.current = current

    def __call__(self) -> datetime:
        return self.current

    def advance(self, **kwargs) -> None:
        self.current += timedelta(**kwargs)


def test_interval_source_immediate_and_next_tick() -> None:
    async def scenario() -> None:
        tz = ZoneInfo("Asia/Shanghai")
        clock = FakeClock(datetime(2026, 3, 7, 10, 0, tzinfo=tz))
        source = IntervalSource.every(
            hours=1,
            immediate=True,
            payload={"job": "sync"},
            timezone="Asia/Shanghai",
            clock=clock,
        )
        await source.open()

        immediate = await source.fetch(1)
        assert len(immediate) == 1
        assert immediate[0].payload == {"job": "sync"}
        assert immediate[0].envelope.meta["scheduled_at"] == "2026-03-07T10:00:00+08:00"
        await immediate[0].ack()

        assert await source.fetch(1) == []
        clock.advance(minutes=59)
        assert await source.fetch(1) == []

        clock.advance(minutes=1)
        hourly = await source.fetch(1)
        assert len(hourly) == 1
        assert hourly[0].envelope.meta["scheduled_at"] == "2026-03-07T11:00:00+08:00"
        await hourly[0].ack()

    asyncio.run(scenario())


def test_interval_source_overlap_skip_drops_missed_ticks() -> None:
    async def scenario() -> None:
        tz = ZoneInfo("Asia/Shanghai")
        clock = FakeClock(datetime(2026, 3, 7, 10, 0, tzinfo=tz))
        source = IntervalSource.every(
            seconds=10,
            overlap="skip",
            timezone=tz,
            clock=clock,
        )
        await source.open()

        clock.advance(seconds=10)
        first = await source.fetch(1)
        assert len(first) == 1
        assert first[0].envelope.meta["scheduled_at"] == "2026-03-07T10:00:10+08:00"

        clock.advance(seconds=25)
        assert await source.fetch(1) == []

        await first[0].ack()
        assert await source.fetch(1) == []

        clock.advance(seconds=5)
        next_run = await source.fetch(1)
        assert len(next_run) == 1
        assert next_run[0].envelope.meta["scheduled_at"] == "2026-03-07T10:00:40+08:00"
        await next_run[0].ack()

    asyncio.run(scenario())


def test_interval_source_overlap_queue_serializes_missed_ticks() -> None:
    async def scenario() -> None:
        tz = ZoneInfo("Asia/Shanghai")
        clock = FakeClock(datetime(2026, 3, 7, 10, 0, tzinfo=tz))
        source = IntervalSource.every(
            seconds=10,
            overlap="queue",
            timezone=tz,
            clock=clock,
        )
        await source.open()

        clock.advance(seconds=10)
        first = await source.fetch(1)
        assert len(first) == 1
        assert first[0].envelope.meta["scheduled_at"] == "2026-03-07T10:00:10+08:00"

        clock.advance(seconds=25)
        assert await source.fetch(1) == []

        await first[0].ack()
        second = await source.fetch(1)
        assert len(second) == 1
        assert second[0].envelope.meta["scheduled_at"] == "2026-03-07T10:00:20+08:00"

        await second[0].ack()
        third = await source.fetch(1)
        assert len(third) == 1
        assert third[0].envelope.meta["scheduled_at"] == "2026-03-07T10:00:30+08:00"
        await third[0].ack()

    asyncio.run(scenario())


def test_cron_source_hourly_alias() -> None:
    async def scenario() -> None:
        tz = ZoneInfo("Asia/Shanghai")
        clock = FakeClock(datetime(2026, 3, 7, 10, 58, tzinfo=tz))
        source = CronSource(
            "@hourly",
            payload={"job": "hourly"},
            timezone="Asia/Shanghai",
            clock=clock,
        )
        await source.open()

        assert await source.fetch(1) == []
        clock.advance(minutes=2)

        first = await source.fetch(1)
        assert len(first) == 1
        assert first[0].payload == {"job": "hourly"}
        assert first[0].envelope.meta["scheduled_at"] == "2026-03-07T11:00:00+08:00"
        await first[0].ack()

        clock.advance(hours=1)
        second = await source.fetch(1)
        assert len(second) == 1
        assert second[0].envelope.meta["scheduled_at"] == "2026-03-07T12:00:00+08:00"
        await second[0].ack()

    asyncio.run(scenario())



import asyncio
import os
import uuid

import pytest

from onestep import SQSConnector


if not os.getenv("ONESTEP_SQS_QUEUE_URL"):
    pytest.skip("set ONESTEP_SQS_QUEUE_URL to run SQS integration tests", allow_module_level=True)


async def _create_queue(connector: SQSConnector, *, fifo: bool = True) -> str:
    client = connector.get_client()
    queue_name = f"onestep-live-{uuid.uuid4().hex}"
    attributes = {}
    if fifo:
        queue_name = f"{queue_name}.fifo"
        attributes = {
            "FifoQueue": "true",
            "ContentBasedDeduplication": "true",
        }
    response = await asyncio.to_thread(client.create_queue, QueueName=queue_name, Attributes=attributes)
    return response["QueueUrl"]


async def _delete_queue(connector: SQSConnector, queue_url: str) -> None:
    await asyncio.to_thread(connector.get_client().delete_queue, QueueUrl=queue_url)


@pytest.mark.integration
def test_sqs_round_trip_live():
    async def scenario():
        region = os.getenv("AWS_REGION") or os.getenv("AWS_DEFAULT_REGION") or "us-east-1"
        connector = SQSConnector(
            region_name=region,
            options={
                "endpoint_url": os.getenv("AWS_ENDPOINT_URL"),
                "aws_access_key_id": os.getenv("AWS_ACCESS_KEY_ID"),
                "aws_secret_access_key": os.getenv("AWS_SECRET_ACCESS_KEY"),
            },
        )
        group_id = os.getenv("ONESTEP_SQS_GROUP_ID")
        queue_url = await _create_queue(connector, fifo=True)
        queue = connector.queue(
            queue_url,
            wait_time_s=1,
            message_group_id=group_id,
            delete_flush_interval_s=0,
        )
        try:
            await queue.publish({"value": 1})
            batch = await queue.fetch(1)
            assert len(batch) == 1
            assert batch[0].payload == {"value": 1}
            await batch[0].ack()
            await queue.close()
        finally:
            await _delete_queue(connector, queue_url)
            await connector.close()

    asyncio.run(scenario())


@pytest.mark.integration
def test_sqs_release_live():
    async def scenario():
        region = os.getenv("AWS_REGION") or os.getenv("AWS_DEFAULT_REGION") or "us-east-1"
        connector = SQSConnector(
            region_name=region,
            options={
                "endpoint_url": os.getenv("AWS_ENDPOINT_URL"),
                "aws_access_key_id": os.getenv("AWS_ACCESS_KEY_ID"),
                "aws_secret_access_key": os.getenv("AWS_SECRET_ACCESS_KEY"),
            },
        )
        queue_url = await _create_queue(connector, fifo=False)
        queue = connector.queue(
            queue_url,
            wait_time_s=1,
            visibility_timeout=2,
            on_fail="release",
            delete_flush_interval_s=0,
        )

        try:
            release_marker = f"release-{uuid.uuid4().hex}"
            await queue.publish({"marker": release_marker, "value": 1})

            first = await queue.fetch(1)
            assert len(first) == 1
            assert first[0].payload["marker"] == release_marker
            await first[0].fail(RuntimeError("release"))

            released = await queue.fetch(1)
            assert len(released) == 1
            assert released[0].payload["marker"] == release_marker
            await released[0].ack()
            empty = await queue.fetch(1)
            assert empty == []

            await queue.close()
        finally:
            await _delete_queue(connector, queue_url)
            await connector.close()

    asyncio.run(scenario())


@pytest.mark.integration
def test_sqs_batch_delete_live():
    async def scenario():
        region = os.getenv("AWS_REGION") or os.getenv("AWS_DEFAULT_REGION") or "us-east-1"
        connector = SQSConnector(
            region_name=region,
            options={
                "endpoint_url": os.getenv("AWS_ENDPOINT_URL"),
                "aws_access_key_id": os.getenv("AWS_ACCESS_KEY_ID"),
                "aws_secret_access_key": os.getenv("AWS_SECRET_ACCESS_KEY"),
            },
        )
        queue_url = await _create_queue(connector, fifo=False)
        queue = connector.queue(
            queue_url,
            wait_time_s=1,
            delete_batch_size=2,
            delete_flush_interval_s=60,
        )

        try:
            batch_marker = f"batch-{uuid.uuid4().hex}"
            await queue.publish({"marker": batch_marker, "value": 2})
            await queue.publish({"marker": batch_marker, "value": 3})

            batch = await queue.fetch(2)
            assert len(batch) == 2
            assert {item.payload["value"] for item in batch} == {2, 3}
            assert all(item.payload["marker"] == batch_marker for item in batch)
            await batch[0].ack()
            await batch[1].ack()

            await queue.flush_deletes()
            empty = await queue.fetch(1)
            assert empty == []

            await queue.close()
        finally:
            await _delete_queue(connector, queue_url)
            await connector.close()

    asyncio.run(scenario())
